# Remove on input neuron from a NN

from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import pandas as pd
import numpy as np
import torch
from torch.utils import data
import pickle
from matplotlib import pyplot as plt
import torch.utils.data as utils
import time
import os

is_cuda = torch.cuda.is_available()

def remove_input_neuron(net,n_inp,idx_neuron,ct_median,save_filename):
    removed_weights = net.linear1.weight[:,idx_neuron]
    # Remove the weights associated with the removed input neuron                                                                                                               
    t = torch.transpose(net.linear1.weight,0,1)
    preserved_ids = torch.LongTensor(np.array(list(set(range(n_inp)) - set([idx_neuron]))))
    t = nn.Parameter(t[preserved_ids, :])
    net.linear1.weight = nn.Parameter(torch.transpose(t,0,1))
    # Adjust the biases
    if is_cuda:
        net.linear1.bias = nn.Parameter(net.linear1.bias+torch.tensor(ct_median*removed_weights).float().cuda())
    else:
        net.linear1.bias = nn.Parameter(net.linear1.bias+torch.tensor(ct_median*removed_weights).float())
    torch.save(net.state_dict(), save_filename)




# runs BF on data and saves the best RPN expressions in results.dat
# all the .dat files are created after I run this script
# the .scr are needed to run the fortran code

import csv
import os
import shutil
import subprocess
import sys
from subprocess import call

import numpy as np
import sympy as sp
from sympy.parsing.sympy_parser import parse_expr

from .resources import _get_resource

# sep_type = 3 for add and 2 for mult and 1 for normal


def brute_force(pathdir, filename, BF_try_time, BF_ops_file_type, sep_type="*", sigma=10, band=0):
    try_time = BF_try_time
    try_time_prefactor = BF_try_time
    file_type = BF_ops_file_type

    try:
        os.remove("results.dat")
    except:
        pass

    try:
        os.remove("brute_solutions.dat")
    except:
        pass

    try:
        os.remove("brute_constant.dat")
    except:
        pass

    try:
        os.remove("brute_formulas.dat")
    except:
        pass

    print("Trying to solve mysteries with brute force...")
    print("Trying to solve {}".format(pathdir+filename))

    shutil.copy2(pathdir+filename, "mystery.dat")

    data = "'{}' '{}' mystery.dat results.dat {:f} {:f}".format(_get_resource(file_type),
                                                                _get_resource(
                                                                    "arity2templates.txt"),
                                                                sigma,
                                                                band)

    with open("args.dat", 'w') as f:
        f.write(data)

    if sep_type == "*":
        try:
            # subprocess.call(["feynman_sr2"], timeout=try_time)
            subprocess.call(["feynman_sr_mdl_mult"], timeout=try_time)
        except:
            pass
    if sep_type == "+":
        try:
            # subprocess.call(["feynman_sr3"], timeout=try_time)
            subprocess.call(["feynman_sr_mdl_plus"], timeout=try_time)
        except:
            pass



# Adds on the pareto all the snapped versions of a given expression (all paramters are snapped in the end)

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as utils
from torch.autograd import Variable
import copy
import warnings
warnings.filterwarnings("ignore")
import sympy
from .S_snap import integerSnap
from .S_snap import zeroSnap
from .S_snap import rationalSnap
from .S_get_symbolic_expr_error import get_symbolic_expr_error
from .get_pareto import Point, ParetoSet

from sympy import preorder_traversal, count_ops
from sympy.abc import x,y
from sympy.parsing.sympy_parser import parse_expr
from sympy import Symbol, lambdify, N, simplify, powsimp, Rational, symbols, S,Float
import re

from .S_get_number_DL_snapped import get_number_DL_snapped

def intify(expr):
    floats = S(expr).atoms(Float)
    ints = [i for i in floats if int(i) == i]
    return expr.xreplace(dict(zip(ints, [int(i) for i in ints])))

# parameters: path to data, math (not RPN) expression
def add_snap_expr_on_pareto(pathdir, filename, math_expr, PA, DR_file=""):
    input_data = np.loadtxt(pathdir+filename)
    def unsnap_recur(expr, param_dict, unsnapped_param_dict):
        """Recursively transform each numerical value into a learnable parameter."""
        import sympy
        from sympy import Symbol
        if isinstance(expr, sympy.numbers.Float) or isinstance(expr, sympy.numbers.Integer) or isinstance(expr, sympy.numbers.Rational) or isinstance(expr, sympy.numbers.Pi):
            used_param_names = list(param_dict.keys()) + list(unsnapped_param_dict)
            unsnapped_param_name = get_next_available_key(used_param_names, "pp", is_underscore=False)
            unsnapped_param_dict[unsnapped_param_name] = float(expr)
            unsnapped_expr = Symbol(unsnapped_param_name)
            return unsnapped_expr
        elif isinstance(expr, sympy.symbol.Symbol):
            return expr
        else:
            unsnapped_sub_expr_list = []
            for sub_expr in expr.args:
                unsnapped_sub_expr = unsnap_recur(sub_expr, param_dict, unsnapped_param_dict)
                unsnapped_sub_expr_list.append(unsnapped_sub_expr)
            return expr.func(*unsnapped_sub_expr_list)


    def get_next_available_key(iterable, key, midfix="", suffix="", is_underscore=True):
        """Get the next available key that does not collide with the keys in the dictionary."""
        if key + suffix not in iterable:
            return key + suffix
        else:
            i = 0
            underscore = "_" if is_underscore else ""
            while "{}{}{}{}{}".format(key, underscore, midfix, i, suffix) in iterable:
                i += 1
            new_key = "{}{}{}{}{}".format(key, underscore, midfix, i, suffix)
            return new_key

    eq = parse_expr(str(math_expr))
    expr = eq

    is_atomic_number = lambda expr:expr.is_Atom and expr.is_number
    eq_numbers = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]

    # Do integer snap one parameter at a time
    integer_snapped_expr = []
    for w in range(len(eq_numbers)):
        try:
            param_dict = {}
            unsnapped_param_dict = {'pp':1}
            eq = unsnap_recur(expr,param_dict,unsnapped_param_dict)
            del unsnapped_param_dict["pp"]
            temp_unsnapped_param_dict = copy.deepcopy(unsnapped_param_dict)
            new_numbers = integerSnap(eq_numbers,w+1)
            new_numbers = {"pp"+str(k): v for k, v in new_numbers.items()}
            temp_unsnapped_param_dict.update(new_numbers)
            #for kk in range(len(new_numbers)):
            #    eq_numbers[new_numbers[kk][0]] = new_numbers[kk][1]
            new_eq = re.sub(r"(pp\d*)",r"{\1}",str(eq))
            new_eq = new_eq.format_map(temp_unsnapped_param_dict)
            integer_snapped_expr = integer_snapped_expr + [parse_expr(new_eq)]
        except:
            continue



    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
    eq_numbers = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]

    # Do rational snap one parameter at a time
    rational_snapped_expr = []
    for w in range(len(eq_numbers)):
        try:
            param_dict = {}
            unsnapped_param_dict = {'pp':1}
            eq = unsnap_recur(expr,param_dict,unsnapped_param_dict)
            del unsnapped_param_dict["pp"]
            temp_unsnapped_param_dict = copy.deepcopy(unsnapped_param_dict)
            new_numbers = rationalSnap(eq_numbers,w+1)
            new_numbers = {"pp"+str(k): v for k, v in new_numbers.items()}
            temp_unsnapped_param_dict.update(new_numbers)
            #for kk in range(len(new_numbers)):
            #    eq_numbers_snap[new_numbers[kk][0]] = new_numbers[kk][1][1:3]
            new_eq = re.sub(r"(pp\d*)",r"{\1}",str(eq))
            new_eq = new_eq.format_map(temp_unsnapped_param_dict)
            rational_snapped_expr = rational_snapped_expr + [parse_expr(new_eq)]
        except:
            continue

    snapped_expr = np.append(integer_snapped_expr,rational_snapped_expr)

    for i in range(len(snapped_expr)):
        try:
            # Calculate the error of the new, snapped expression
            snapped_error = get_symbolic_expr_error(input_data,str(snapped_expr[i]))
            # Calculate the complexity of the new, snapped expression
            #expr = simplify(powsimp(snapped_expr[i]))
            expr = snapped_expr[i]
            for s in (expr.free_symbols):
                s = symbols(str(s), real = True)
            expr =  parse_expr(str(snapped_expr[i]),locals())
            expr = intify(expr)
            is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
            numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]

            if DR_file=="":
                snapped_complexity = 0
                for j in numbers_expr:
                    snapped_complexity = snapped_complexity + get_number_DL_snapped(float(j))

                n_variables = len(expr.free_symbols)
                n_operations = len(count_ops(expr,visual=True).free_symbols)
                if n_operations!=0 or n_variables!=0:
                    snapped_complexity = snapped_complexity + (n_variables+n_operations)*np.log2((n_variables+n_operations))

            # If a da file is provided, replace the variables with the actual ones before calculating the complexity
            else:
                dr_data = np.loadtxt(DR_file,dtype="str",delimiter=",")

                expr = str(expr)
                old_vars = ["x%s" %k for k in range(len(dr_data)-3)]
                for i_dr in range(len(old_vars)):
                    expr = expr.replace(old_vars[i_dr],"("+dr_data[i_dr+2]+")")
                expr = "("+dr_data[1]+")*(" + expr +")"

                expr = parse_expr(expr)
                for s in (expr.free_symbols):
                    s = symbols(str(s), real = True)
                #expr =  simplify(parse_expr(str(expr),locals()))
                expr =  parse_expr(str(expr),locals())
                snapped_complexity = 0
                for j in numbers_expr:
                    snapped_complexity = snapped_complexity + get_number_DL_snapped(float(j))

                n_variables = len(expr.free_symbols)
                n_operations = len(count_ops(expr,visual=True).free_symbols)
                if n_operations!=0 or n_variables!=0:
                    snapped_complexity = snapped_complexity + (n_variables+n_operations)*np.log2((n_variables+n_operations))

            PA.add(Point(x=snapped_complexity, y=snapped_error, data=str(expr)))
        except:
            continue
    return(PA)







# Turns a mathematical expression (already RPN turned) to pytorch expression, trains the parameters, and returns the new error, complexity and the new symbolic expression

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as utils
from torch.autograd import Variable
import warnings
warnings.filterwarnings("ignore")
import sympy

from sympy import *
from sympy.abc import x,y
from sympy.parsing.sympy_parser import parse_expr
from sympy import Symbol, lambdify, N

from .S_get_number_DL_snapped import get_number_DL_snapped
from .S_get_symbolic_expr_error import get_symbolic_expr_error

# parameters: path to data, RPN expression (obtained from bf)
def RPN_to_pytorch(data, math_expr, lr = 1e-2, N_epochs = 500):
    param_dict = {}
    unsnapped_param_dict = {'p':1}

    def unsnap_recur(expr, param_dict, unsnapped_param_dict):
        """Recursively transform each numerical value into a learnable parameter."""
        import sympy
        from sympy import Symbol
        if isinstance(expr, sympy.numbers.Float) or isinstance(expr, sympy.numbers.Integer) or isinstance(expr, sympy.numbers.Rational) or isinstance(expr, sympy.numbers.Pi):
            used_param_names = list(param_dict.keys()) + list(unsnapped_param_dict)
            unsnapped_param_name = get_next_available_key(used_param_names, "p", is_underscore=False)
            unsnapped_param_dict[unsnapped_param_name] = float(expr)
            unsnapped_expr = Symbol(unsnapped_param_name)
            return unsnapped_expr
        elif isinstance(expr, sympy.symbol.Symbol):
            return expr
        else:
            unsnapped_sub_expr_list = []
            for sub_expr in expr.args:
                unsnapped_sub_expr = unsnap_recur(sub_expr, param_dict, unsnapped_param_dict)
                unsnapped_sub_expr_list.append(unsnapped_sub_expr)
            return expr.func(*unsnapped_sub_expr_list)


    def get_next_available_key(iterable, key, midfix="", suffix="", is_underscore=True):
        """Get the next available key that does not collide with the keys in the dictionary."""
        if key + suffix not in iterable:
            return key + suffix
        else:
            i = 0
            underscore = "_" if is_underscore else ""
            while "{}{}{}{}{}".format(key, underscore, midfix, i, suffix) in iterable:
                i += 1
            new_key = "{}{}{}{}{}".format(key, underscore, midfix, i, suffix)
            return new_key

    # Turn BF expression to pytorch expression
    eq = parse_expr(math_expr)
    eq = unsnap_recur(eq,param_dict,unsnapped_param_dict)

    N_vars = len(data[0])-1
    N_params = len(unsnapped_param_dict)

    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    variables = []
    params = []
    for i in range(N_vars):
        variables = variables + [possible_vars[i]]
    for i in range(N_params-1):
        params = params + ["p%s" %i]

    symbols = params + variables

    f = lambdify(symbols, N(eq), torch)

    # Set the trainable parameters in the expression

    trainable_parameters = []
    for i in unsnapped_param_dict:
        if i!="p":
            vars()[i] = torch.tensor(unsnapped_param_dict[i])
            vars()[i].requires_grad=True
            trainable_parameters = trainable_parameters + [vars()[i]]

    # Prepare the loaded data
    real_variables = []
    for i in range(len(data[0])-1):
        real_variables = real_variables + [torch.from_numpy(data[:,i]).float()]

    input = trainable_parameters + real_variables
    y = torch.from_numpy(data[:,-1]).float()

    for i in range(N_epochs):
        # this order is fixed i.e. first parameters
        yy = f(*input)
        loss = torch.mean((yy-y)**2)
        loss.backward()
        with torch.no_grad():
            for j in range(N_params-1):
                trainable_parameters[j] -= lr * trainable_parameters[j].grad
                trainable_parameters[j].grad.zero_()
        if torch.isnan(loss):
            break

    for nan_i in range(len(trainable_parameters)):
        if torch.isnan(trainable_parameters[nan_i])==True or abs(trainable_parameters[nan_i])>1e7:
            return 1000000, 10000000, "1"

    ii = -1
    for parm in unsnapped_param_dict:
        if ii == -1:
            ii = ii + 1
        else:
            eq = eq.subs(parm, trainable_parameters[ii])
            ii = ii + 1

    complexity = 0
    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
    numbers_expr = [subexpression for subexpression in preorder_traversal(eq) if is_atomic_number(subexpression)]
    complexity = 0
    for j in numbers_expr:
        try:
            complexity = complexity + get_number_DL_snapped(float(j))
        except:
            complexity = complexity + 1000000
    n_variables = len(eq.free_symbols)
    n_operations = len(count_ops(eq,visual=True).free_symbols)
    if n_operations!=0 or n_variables!=0:
        complexity = complexity + (n_variables+n_operations)*np.log2((n_variables+n_operations))

    error = get_symbolic_expr_error(data,str(eq))
    return error, complexity, eq







"""AI Feynman
"""


from .S_run_aifeynman import run_aifeynman
from .get_demos import get_demos

__title__ = "aifeynman"
__version__ = "2.0.0"
__license__ = "MIT"
__copyright__ = "Copyright 2020 Silviu-Marian Udrescu"

__all__ = ["run_aifeynman"]



from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
from sortedcontainers import SortedKeyList


class Point(object):
    def __init__(self, x, y, data=None, id=None):
        self.x = x
        self.y = y
        self.data = data
        self.id = id


    def __getitem__(self, index):
        """Indexing: get item according to index."""
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.data
        elif index == 3:
            return self.id
        else:
            raise Exception("Index {} is out of range!".format(index))


    def __setitem__(self, index, value):
        """Indexing: set item according to index."""
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.data = value
        elif index == 3:
            raise Exception("Cannot set Id!")
        else:
            raise Exception("Index {} is out of range!".format(index))


# In[2]:


class ParetoSet(SortedKeyList):
    """Maintained maximal set with efficient insertion. Note that we use the convention of smaller the better."""

    def __init__(self):
        super().__init__(key=lambda p: p.x)


    def _input_check(self, p):
        """Check that input is in the correct format.

        Args:
            p: input

        Returns:
            Point:

        Raises:
            TypeError if cannot be converted.
        """

        if isinstance(p, Point):
            return p
        elif isinstance(p, tuple) and len(p) == 2:
            return Point(x=p[0], y=p[1], data=None)
        else:
            raise TypeError("Must be instance of Point or 2-tuple.")
    
    
    def get_id_list(self):
        id_list = []
        for point in self:
            id_list.append(point.id)
        return id_list

    
    def add(self, p):
        """Insert Point into set if minimal in first two indices.

        Args:
            p (Point): Point to insert

        Returns:
            bool: True only if point is inserted

        """
        p = self._input_check(p)

        is_pareto = False
        # check right for dominated points:
        right = self.bisect_left(p)

        while len(self) > right and self[right].y >= p.y and not (self[right].x == p.x and self[right].y == p.y):
            self.pop(right)
            is_pareto = True

        # check left for dominating points:
        left = self.bisect_right(p) - 1

        if left == -1 or self[left][1] > p[1]:
            is_pareto = True

        # if it's the only point it's maximal
        if len(self) == 0:
            is_pareto = True

        if is_pareto:
            super().add(p)

        return is_pareto


    def __contains__(self, p):
        p = self._input_check(p)

        left = self.bisect_left(p)

        while len(self) > left and self[left].x == p.x:
            if self[left].y == p.y:
                return True

            left += 1

        return False


    def __add__(self, other):
        """Merge another pareto set into self.

        Args:
            other (ParetoSet): set to merge into self

        Returns:
            ParetoSet: self

        """

        for item in other:
            self.add(item)

        return self


    def distance(self, p):
        """Given a Point, calculate the minimum Euclidean distance to pareto
        frontier (in first two indices).

        Args:
            p (Point): point

        Returns:
            float: minimum Euclidean distance to pareto frontier

        """
        p = self._input_check(p)

        point = np.array((p.x, p.y))
        dom = self.dominant_array(p)

        # distance is zero if pareto optimal
        if dom.shape[0] == 0:
            return 0.

        # add corners of all adjacent pairs
        candidates = np.zeros((dom.shape[0] + 1, 2))
        for i in range(dom.shape[0] - 1):
            candidates[i, :] = np.max(dom[[i, i+1], :], axis=0)

        # add top and right bounds
        candidates[-1, :] = (p.x, np.min(dom[:, 1]))
        candidates[-2, :] = (np.min(dom[:, 0]), p.y)

        return np.min(np.sqrt(np.sum(np.square(candidates - point), axis=1)))


    def dominant_array(self, p):
        """Given a Point, return the set of dominating points in the set (in
        the first two indices).

        Args:
            p (Point): point

        Returns:
            numpy.ndarray: array of dominating points

        """
        p = self._input_check(p)

        idx = self.bisect_left(p) - 1

        domlist = []

        while idx >= 0 and self[idx][1] < p[1]:
            domlist.append(self[idx])
            idx -= 1

        return np.array([x[0:2] for x in domlist])


    def to_array(self):
        """Convert first two indices to numpy.ndarray

        Args:
            None

        Returns:
            numpy.ndarray: array of shape (len(self), 2)

        """
        A = np.zeros((len(self), 2))
        for i, p in enumerate(self):
            A[i, :] = p.x, p.y

        return A

    def get_pareto_points(self):
        """Returns the x, y and data for each point in the pareto frontier
        
        """
        pareto_points = []
        for i, p in enumerate(self):
            pareto_points = pareto_points + [[p.x, p.y, p.data]]
        
        return pareto_points
        

    def from_list(self, A):
        """Convert iterable of Points into ParetoSet.

        Args:
            A (iterator): iterator of Points

        Returns:
            None

        """
        for a in A:
            self.add(a)
    
    
    def plot(self):
        """Plotting the Pareto frontier."""
        array = self.to_array()
        plt.figure(figsize=(8, 6))
        plt.plot(array[:, 0], array[:, 1], 'r.')
        plt.show()


if __name__ == "__main__":
    PA = ParetoSet()
    A = np.zeros((40, 2))
    
    for i in range(40):
        x = np.random.rand()
        y = np.random.rand()
        
        A[i, 0] = x
        A[i, 1] = y
        
        PA.add(Point(x=x, y=y, data=None))
    paretoA = PA.to_array()




# Turns an RPN expression to normal mathematical notation

import numpy as np

def RPN_to_eq(expr):

    variables = ["0","1","a","b","c","d","e","f","g","h","i","j","k","l","m","n","P"]
    operations_1 = [">","<","~","\\","L","E","S","C","A","N","T","R","O","J"]
    operations_2 = ["+","*","-","/"]

    stack = np.array([])

    for i in (expr):
        if i in variables:
            if i == "P":
                stack = np.append(stack,"pi")
            elif i == "0":
                stack = np.append(stack,"0")
            elif i == "1":
                stack = np.append(stack,"1")  
            else:
                stack = np.append(stack,"x" + str(ord(i)-97))
        elif i in operations_2:
            a1 = stack[-1]
            a2 = stack[-2]
            stack = np.delete(stack,-1)
            stack = np.delete(stack,-1)
            a = "("+a2+i+a1+")"
            stack = np.append(stack,a)
        elif i in operations_1:
            a = stack[-1]
            stack = np.delete(stack,-1)
            if i==">":
                a="("+a+"+1)"
                stack = np.append(stack,a)
            if i=="<":
                a="("+a+"-1)"
                stack = np.append(stack,a)
            if i=="~":
                a="(-"+a+")"
                stack = np.append(stack,a)
            if i=="\\":
                a="("+a+")**(-1)"
                stack = np.append(stack,a)
            if i=="L":
                a="log("+a+")"
                stack = np.append(stack,a)
            if i=="E":
                a="exp("+a+")"
                stack = np.append(stack,a)
            if i=="S":
                a="sin("+a+")"
                stack = np.append(stack,a)
            if i=="C":
                a="cos("+a+")"
                stack = np.append(stack,a)
            if i=="A":
                a="abs("+a+")"
                stack = np.append(stack,a)
            if i=="N":
                a="asin("+a+")"
                stack = np.append(stack,a)
            if i=="T":
                a="atan("+a+")"
                stack = np.append(stack,a)
            if i=="R":
                a="sqrt("+a+")"
                stack = np.append(stack,a)
            if i=="O":
                a="(2*("+a+"))"
                stack = np.append(stack,a)
            if i=="J":
                a="(2*("+a+")+1)"
                stack = np.append(stack,a)
    return(stack[0])



# Adds on the pareto all the snapped versions of a given expression (all paramters are snapped in the end)

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as utils
from torch.autograd import Variable
import copy
import warnings
warnings.filterwarnings("ignore")
import sympy
from .S_snap import integerSnap
from .S_snap import zeroSnap
from .S_snap import rationalSnap
from .S_get_symbolic_expr_error import get_symbolic_expr_error
from .get_pareto import Point, ParetoSet
from .S_brute_force_number import brute_force_number

from sympy import preorder_traversal, count_ops
from sympy.abc import x,y
from sympy.parsing.sympy_parser import parse_expr
from sympy import Symbol, lambdify, N, simplify, powsimp
from .RPN_to_eq import RPN_to_eq

from .S_get_number_DL_snapped import get_number_DL_snapped

# parameters: path to data, math (not RPN) expression
def add_bf_on_numbers_on_pareto(pathdir, filename, PA, math_expr):
    input_data = np.loadtxt(pathdir+filename)
    def unsnap_recur(expr, param_dict, unsnapped_param_dict):
        """Recursively transform each numerical value into a learnable parameter."""
        import sympy
        from sympy import Symbol
        if isinstance(expr, sympy.numbers.Float) or isinstance(expr, sympy.numbers.Integer) or isinstance(expr, sympy.numbers.Rational) or isinstance(expr, sympy.numbers.Pi):
            used_param_names = list(param_dict.keys()) + list(unsnapped_param_dict)
            unsnapped_param_name = get_next_available_key(used_param_names, "p", is_underscore=False)
            unsnapped_param_dict[unsnapped_param_name] = float(expr)
            unsnapped_expr = Symbol(unsnapped_param_name)
            return unsnapped_expr
        elif isinstance(expr, sympy.symbol.Symbol):
            return expr
        else:
            unsnapped_sub_expr_list = []
            for sub_expr in expr.args:
                unsnapped_sub_expr = unsnap_recur(sub_expr, param_dict, unsnapped_param_dict)
                unsnapped_sub_expr_list.append(unsnapped_sub_expr)
            return expr.func(*unsnapped_sub_expr_list)


    def get_next_available_key(iterable, key, midfix="", suffix="", is_underscore=True):
        """Get the next available key that does not collide with the keys in the dictionary."""
        if key + suffix not in iterable:
            return key + suffix
        else:
            i = 0
            underscore = "_" if is_underscore else ""
            while "{}{}{}{}{}".format(key, underscore, midfix, i, suffix) in iterable:
                i += 1
            new_key = "{}{}{}{}{}".format(key, underscore, midfix, i, suffix)
            return new_key

    eq = parse_expr(str(math_expr))
    expr = eq
    # Get the numbers appearing in the expression
    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
    eq_numbers = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]
    # Do bf on one parameter at a time
    bf_on_numbers_expr = []
    for w in range(len(eq_numbers)):
        try:
            param_dict = {}
            unsnapped_param_dict = {'p':1}
            eq_ = unsnap_recur(expr,param_dict,unsnapped_param_dict)
            eq = eq_

            np.savetxt(pathdir+"number_for_bf_%s.txt" %w, [eq_numbers[w]])
            brute_force_number(pathdir,"number_for_bf_%s.txt" %w)
            # Load the predictions made by the bf code
            bf_numbers = np.loadtxt("results.dat",usecols=(1,),dtype="str")
            new_numbers = copy.deepcopy(eq_numbers)

            # replace the number under consideration by all the proposed bf numbers
            for kk in range(len(bf_numbers)):
                eq = eq_
                new_numbers[w] = parse_expr(RPN_to_eq(bf_numbers[kk]))

                jj = 0
                for parm in unsnapped_param_dict:
                    if parm!="p":
                        eq = eq.subs(parm, new_numbers[jj])
                        jj = jj + 1

                bf_on_numbers_expr = bf_on_numbers_expr + [eq]
        except:
            continue

    for i in range(len(bf_on_numbers_expr)):
        try:
            # Calculate the error of the new, snapped expression
            snapped_error = get_symbolic_expr_error(input_data,str(bf_on_numbers_expr[i]))
            # Calculate the complexity of the new, snapped expression
            expr = simplify(powsimp(bf_on_numbers_expr[i]))
            is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
            numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]

            snapped_complexity = 0
            for j in numbers_expr:
                snapped_complexity = snapped_complexity + get_number_DL_snapped(float(j))
            # Add the complexity due to symbols
            n_variables = len(expr.free_symbols)
            n_operations = len(count_ops(expr,visual=True).free_symbols)
            if n_operations!=0 or n_variables!=0:
                snapped_complexity = snapped_complexity + (n_variables+n_operations)*np.log2((n_variables+n_operations))

            PA.add(Point(x=snapped_complexity, y=snapped_error, data=str(expr)))
        except:
            continue

    return(PA)




# Calculates the error of a given symbolic expression applied to a dataset. The input should be a string of the mathematical expression

from .get_pareto import Point, ParetoSet
from sympy.parsing.sympy_parser import parse_expr
import numpy as np
import matplotlib.pyplot as plt
import os
from os import path
from sympy import Symbol, lambdify, N

def get_symbolic_expr_error(data,expr):
    try:
        N_vars = len(data[0])-1
        possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
        variables = []
        for i in range(N_vars):
            variables = variables + [possible_vars[i]]
        eq = parse_expr(expr)
        f = lambdify(variables, N(eq))
        real_variables = []

        for i in range(len(data[0])-1):
            check_var = "x"+str(i)
            if check_var in np.array(variables).astype('str'):
                real_variables = real_variables + [data[:,i]]

        # Remove accidental nan's
        good_idx = np.where(np.isnan(f(*real_variables))==False)

        # use this to get rid of cases where the loss gets complex because of transformations of the output variable
        if isinstance(np.mean((f(*real_variables)-data[:,-1])**2), complex):
            return 1000000
        else:
            try:
                return np.mean(np.log2(1+abs(f(*real_variables)[good_idx]-data[good_idx][:,-1])*2**30))
            except:
                return np.mean(np.log2(1+abs(f(*real_variables)-data[:,-1])*2**30))
    except:
        return 1000000



import numpy as np
import os
from .S_polyfit_utils import getBest
from .S_polyfit_utils import basis_vector
import itertools
import sys
import csv
import sympy
from sympy import symbols, Add, Mul, S, simplify
from scipy.linalg import fractional_matrix_power

def mk_sympy_function(coeffs, num_covariates, deg):
    generators = [basis_vector(num_covariates+1, i) for i in range(num_covariates+1)]
    powers = map(sum, itertools.combinations_with_replacement(generators, deg))

    coeffs = np.round(coeffs,2)

    xs = (S.One,) + symbols('z0:%d'%num_covariates)
    if len(coeffs)>1:
        return Add(*[coeff * Mul(*[x**deg for x, deg in zip(xs, power)])
                     for power, coeff in zip(powers, coeffs)])
    else:
        return coeffs[0]

def polyfit(maxdeg, filename):
    n_variables = np.loadtxt(filename, dtype='str').shape[1]-1
    variables = np.loadtxt(filename, usecols=(0,))
    means = [np.mean(variables)]

    for j in range(1,n_variables):
        v = np.loadtxt(filename, usecols=(j,))
        means = means + [np.mean(v)]
        variables = np.column_stack((variables,v))

    f_dependent = np.loadtxt(filename, usecols=(n_variables,))

    if n_variables>1:
        C_1_2 = fractional_matrix_power(np.cov(variables.T),-1/2)
        x = []
        z = []
        for ii in range(len(variables[0])):
            variables[:,ii] = variables[:,ii] - np.mean(variables[:,ii])
            x = x + ["x"+str(ii)]
            z = z + ["z"+str(ii)]

        if np.isnan(C_1_2).any()==False:
            variables = np.matmul(C_1_2,variables.T).T
            res = getBest(variables,f_dependent,maxdeg)
            parameters = res[0]
            params_error = res[1]
            deg = res[2]

            x = sympy.Matrix(x)
            M = sympy.Matrix(C_1_2)
            b = sympy.Matrix(means)
            M_x = M*(x-b)

            eq = mk_sympy_function(parameters,n_variables,deg)
            symb = sympy.Matrix(z)

            for i in range(len(symb)):
                eq = eq.subs(symb[i],M_x[i])

            eq = simplify(eq)

        else:
            res = getBest(variables,f_dependent,maxdeg)
            parameters = res[0]
            params_error = res[1]
            deg = res[2]

            eq = mk_sympy_function(parameters,n_variables,deg)
            for i in range(len(x)):
                eq = eq.subs(z[i],x[i])
            eq = simplify(eq)

    else:
        res = getBest(variables,f_dependent,maxdeg)
        parameters = res[0]
        params_error = res[1]
        deg = res[2]
        eq = mk_sympy_function(parameters,n_variables,deg)
        try:
            eq = eq.subs("z0","x0")
        except:
            pass

    return (eq, params_error)




# Combines 2 pareto fromtier obtained from the separability test into a new one.

from .get_pareto import Point, ParetoSet
from sympy.parsing.sympy_parser import parse_expr
import numpy as np
import matplotlib.pyplot as plt
import os
from os import path
from sympy import Symbol, lambdify, N
from .get_pareto import Point, ParetoSet
from .S_get_expr_complexity import get_expr_complexity

def add_sym_on_pareto(pathdir,filename,PA1,idx1,idx2,PA,sym_typ):
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    PA1 = np.array(PA1.get_pareto_points()).astype('str')
    for i in range(len(PA1)):
        exp1 = PA1[i][2]
        for j in range(len(possible_vars)-2,idx2-1,-1):
            exp1 = exp1.replace(possible_vars[j],possible_vars[j+1])
        exp1 = exp1.replace(possible_vars[idx1],"(" + possible_vars[idx1] + sym_typ + possible_vars[idx2] + ")")
        compl = get_expr_complexity(exp1)
        PA.add(Point(x=compl,y=float(PA1[i][1]),data=str(exp1)))

    return PA







import numpy as np
from .RPN_to_eq import RPN_to_eq
from scipy.optimize import fsolve
from sympy import lambdify, N
import torch
import torch.nn as nn
import torch.nn.functional as F
from .get_pareto import Point, ParetoSet
from .S_get_expr_complexity import get_expr_complexity
from . import test_points
import os
import warnings
warnings.filterwarnings("ignore")
is_cuda = torch.cuda.is_available()


def check_compositionality(pathdir,filename,model,express,mu,sigma,nu=10):
    data = np.loadtxt(pathdir+filename)

    eq = RPN_to_eq(express)
    print(eq,express)
    # Get the variables appearing in the equation
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    variables = []

    N_vars = len(data[0])-1
    for i in range(N_vars):
        variables = variables + [possible_vars[i]]
    symbols = variables
    f = lambdify(symbols, N(eq))

    fixed = data[:,0:-1]
    length_fixed = len(fixed)

    bm = np.ones(len(data[0])-1,dtype=bool)
    obj = test_points.init_general_test_point(eq, data[:,:-1], data[:,-1], bm)

    list_z = np.array([])
    z = 0
    i = 0
    model.eval()
    while z<nu and i<len(data[0:1000]):
        # Generate functions based on the discovered possible equation and check if they are right
        dt = test_points.get_test_point(obj,data[i][:-1])
        diff = abs(f(*fixed[i])-f(*dt))
        with torch.no_grad():
            if diff<1e-4:
                if is_cuda:
                    dt = torch.tensor(dt).float().cuda().view(1,len(dt))
                    dt = torch.cat((torch.tensor([np.zeros(len(dt[0]))]).float().cuda(),dt), 0)
                    error = torch.tensor(data[:,-1][i]).cuda()-model(dt)[1:]
                    error = error.cpu().detach().numpy()
                    list_z = np.append(list_z,np.log2(1+abs(error)*2**30))
                    z = np.sqrt(len(list_z))*(np.mean(list_z)-mu)/sigma
                else:
                    dt = torch.tensor(dt).float().view(1,len(dt))
                    dt = torch.cat((torch.tensor([np.zeros(len(dt[0]))]).float(),dt), 0)
                    error =torch.tensor(data[:,-1][i])-model(dt)[1:]
                    error = error.detach().numpy()
                    list_z = np.append(list_z,np.log2(1+abs(error)*2**30))
                    z = np.sqrt(len(list_z))*(np.mean(list_z)-mu)/sigma
                i = i + 1
            else:
                i = i + 1

    print(i)

    if i==len(data[0:1000]) and np.mean(list_z)<mu:
        print("return 1!!!",1,express,np.mean(list_z),np.std(list_z))
        return (1,express,np.mean(list_z),np.std(list_z))
    else:
        return (0,express,100,100)

#print(check_compositionality("../","table_data_8.txt",model,"dbacd/+~+/",50,2))

def do_compositionality(pathdir,filename,express):
    data = np.loadtxt(pathdir+filename)
    eq = RPN_to_eq(express)
    # Get the variables appearing in the equation
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    variables = []

    N_vars = len(data[0])-1
    for i in range(N_vars):
        variables = variables + [possible_vars[i]]

    symbols = variables
    f = lambdify(symbols, N(eq))

    new_data = f(*np.transpose(data[:,0:-1]))
    save_data = np.column_stack((new_data,data[:,-1]))

    try:
        os.mkdir("results/compositionality")
    except:
        pass

    file_name = filename + "-comp"
    np.savetxt("results/compositionality/"+file_name,save_data)

    return ("results/compositionality/", file_name)


def add_comp_on_pareto(PA1,PA,express):
    eq = RPN_to_eq(express)
    PA1 = np.array(PA1.get_pareto_points()).astype('str')
    for i in range(len(PA1)):
        exp1 = PA1[i][2]
        print(exp1)
        exp1 = exp1.replace("x0",eq)
        compl = get_expr_complexity(exp1)
        PA.add(Point(x=compl,y=float(PA1[i][1]),data=str(exp1)))

    return PA




import numpy as np
import pandas as pd
from scipy.sparse.linalg import lsqr
from scipy.linalg import *
from sympy import Matrix
from sympy import symbols, Add, Mul, S
from numpy.linalg import matrix_rank
from itertools import combinations


N = np.array([[ 0,  1,  1],
              [ 0, -1, -1],
              [ 1,  0, 0],
              [ 0,  0,  0],
              [ 0,  0,  0],
              [ 0,  0, 0],])

N = np.array([[ 0,  0,  3,  1,  1,  1,  1,  1,  1],
              [ 0,  0, -2,  0,  0,  0,  0,  0,  0],
              [ 1,  1, -1,  0,  0,  0,  0,  0,  0],
              [ 0,  0,  0,  0,  0,  0,  0,  0,  0],
              [ 0,  0,  0,  0,  0,  0,  0,  0,  0],
              [ 0,  0,  0,  0,  0,  0,  0,  0,  0]])

a = np.array([ 1, -2,  1,  0,  0,  0])

def getPowers(N,a):
    rand_drop_cols = np.arange(0,len(N[0]),1)
    rand_drop_rows = np.arange(0,len(N),1)
    rand_drop_rows = np.flip(rand_drop_rows)
    rank = matrix_rank(N)
    d_cols = list(combinations(rand_drop_cols,len(N[0])-rank))
    d_rows = list(combinations(rand_drop_rows,len(N)-rank))
    for i in d_cols:
        M = N
        M = np.delete(M,i,1)
        M = np.transpose(M)
        for j in d_rows:
            P = M
            P = np.delete(P,j,1)
            if np.linalg.det(P)!=0:
                solved_M = np.transpose(P)
                indices_sol = j
                indices_powers = i
                break

    b = np.delete(a,indices_sol)
    params = np.linalg.solve(solved_M,b)

    sol = []
    for i in range(len(N[0])):
        if i in indices_powers:
            sol = sol + [0]
        else:
            sol = sol + [params[0]]
            params = np.delete(params,0)

    # this is the solution:
    sol = np.array(sol)
    return(sol)





# checks for symmetries in the data

from __future__ import print_function
import torch
import os
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import pandas as pd
import numpy as np
import torch
from torch.utils import data
import pickle
from torch.optim.lr_scheduler import CosineAnnealingLR
from matplotlib import pyplot as plt
from .S_remove_input_neuron import remove_input_neuron
import time

is_cuda = torch.cuda.is_available()

class SimpleNet(nn.Module):
    def __init__(self, ni):
        super().__init__()
        self.linear1 = nn.Linear(ni, 128)
        self.linear2 = nn.Linear(128, 128)
        self.linear3 = nn.Linear(128, 64)
        self.linear4 = nn.Linear(64,64)
        self.linear5 = nn.Linear(64,1)

    def forward(self, x):
        x = F.tanh(self.linear1(x))
        x = F.tanh(self.linear2(x))
        x = F.tanh(self.linear3(x))
        x = F.tanh(self.linear4(x))
        x = self.linear5(x)
        return x

def rmse_loss(pred, targ):
    denom = targ**2
    denom = torch.sqrt(denom.sum()/len(denom))
    return torch.sqrt(F.mse_loss(pred, targ))/denom

# checks if f(x,y)=f(x-y)
def check_translational_symmetry_minus(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD \n")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+"/%s" %filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():            
            # make the shift x->x+a for 2 variables at a time (different variables)
            min_error = 1000
            best_i = -1
            best_j = -1
            best_mu = 0
            best_sigma = 0
            for i in range(0,n_variables,1):
                for j in range(0,n_variables,1):
                    if i<j:
                        fact_translate = factors.clone()
                        a = 0.5*min(torch.std(fact_translate[:,i]),torch.std(fact_translate[:,j]))
                        fact_translate[:,i] = fact_translate[:,i] + a
                        fact_translate[:,j] = fact_translate[:,j] + a
                        list_errs = abs(product-model(fact_translate))
                        error = torch.median(list_errs)
                        mu = torch.mean(torch.log2(1+list_errs*2**30))
                        sigma = torch.std(torch.log2(1+list_errs*2**30))
                        if error<min_error:
                            min_error = error
                            best_i = i
                            best_j = j
                            best_mu = mu
                            best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma

    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)
    
def do_translational_symmetry_minus(pathdir, filename, i,j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        for k in range(1,n_variables):
            v = np.loadtxt(pathdir+"/%s" %filename, usecols=(k,))
            variables = np.column_stack((variables,v))
        
        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():   
            file_name = filename + "-translated_minus"
            ct_median = torch.median(torch.from_numpy(variables[:,j]))
            data_translated = variables
            data_translated[:,i] = variables[:,i]-variables[:,j]
            data_translated =  np.delete(data_translated, j, axis=1)
            data_translated = np.column_stack((data_translated,f_dependent))
            try:
                os.mkdir("results/translated_data_minus/")
            except:
                pass
            np.savetxt("results/translated_data_minus/"+file_name , data_translated)
            remove_input_neuron(model,n_variables,j,ct_median,"results/NN_trained_models/models/"+filename + "-translated_minus_pretrained.h5")
            return ("results/translated_data_minus/",file_name)

    except Exception as e:
        print(e)
        return (-1,-1)
        

# checks if f(x,y)=f(x/y)
def check_translational_symmetry_divide(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD \n")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+"/%s" %filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():
            a = 1.2
            min_error = 1000
            best_i = -1
            best_j = -1
            best_mu = 0
            best_sigma = 0
            # make the shift x->x*a and y->y*a for 2 variables at a time (different variables)
            for i in range(0,n_variables,1):
                for j in range(0,n_variables,1):
                    if i<j:
                        fact_translate = factors.clone()
                        fact_translate[:,i] = fact_translate[:,i]*a
                        fact_translate[:,j] = fact_translate[:,j]*a
                        list_errs = abs(product-model(fact_translate))
                        error = torch.median(list_errs)
                        mu = torch.mean(torch.log2(1+list_errs*2**30))
                        sigma = torch.std(torch.log2(1+list_errs*2**30))
                        if error<min_error:
                            min_error = error
                            best_i = i
                            best_j = j
                            best_mu = mu
                            best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma

    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)
    
    
def do_translational_symmetry_divide(pathdir, filename, i,j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        for k in range(1,n_variables):
            v = np.loadtxt(pathdir+"/%s" %filename, usecols=(k,))
            variables = np.column_stack((variables,v))
        
        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():             
            file_name = filename + "-translated_divide"
            data_translated = variables
            ct_median =torch.median(torch.from_numpy(variables[:,j]))
            data_translated[:,i] = variables[:,i]/variables[:,j]
            data_translated =  np.delete(data_translated, j, axis=1)
            data_translated = np.column_stack((data_translated,f_dependent))
            try:
                os.mkdir("results/translated_data_divide/")
            except:
                pass
            np.savetxt("results/translated_data_divide/"+file_name , data_translated)
            remove_input_neuron(model,n_variables,j,ct_median,"results/NN_trained_models/models/"+filename + "-translated_divide_pretrained.h5")
            return ("results/translated_data_divide/",file_name)

    except Exception as e:
        print(e)
        return (-1,1)

# checks if f(x,y)=f(x*y)
def check_translational_symmetry_multiply(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD \n")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+"/%s" %filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():
            a = 1.2
            min_error = 1000
            best_i = -1
            best_j = -1
            best_mu = 0
            best_sigma = 0
            # make the shift x->x*a and y->y/a for 2 variables at a time (different variables)
            for i in range(0,n_variables,1):
                for j in range(0,n_variables,1):
                    if i<j:
                        fact_translate = factors.clone()
                        fact_translate[:,i] = fact_translate[:,i]*a
                        fact_translate[:,j] = fact_translate[:,j]/a
                        list_errs = abs(product-model(fact_translate))
                        error = torch.median(list_errs)
                        mu = torch.mean(torch.log2(1+list_errs*2**30))
                        sigma = torch.std(torch.log2(1+list_errs*2**30))
                        if error<min_error:
                            min_error = error
                            best_i = i
                            best_j = j
                            best_mu = mu
                            best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma

    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)

def do_translational_symmetry_multiply(pathdir, filename, i,j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        for k in range(1,n_variables):
            v = np.loadtxt(pathdir+"/%s" %filename, usecols=(k,))
            variables = np.column_stack((variables,v))
        
        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():             
            file_name = filename + "-translated_multiply"
            data_translated = variables
            ct_median =torch.median(torch.from_numpy(variables[:,j]))
            data_translated[:,i] = variables[:,i]*variables[:,j]
            data_translated =  np.delete(data_translated, j, axis=1)
            data_translated = np.column_stack((data_translated,f_dependent))
            try:
                os.mkdir("results/translated_data_multiply/")
            except:
                pass
            np.savetxt("results/translated_data_multiply/"+file_name , data_translated)
            remove_input_neuron(model,n_variables,j,ct_median,"results/NN_trained_models/models/"+filename + "-translated_multiply_pretrained.h5")
            return ("results/translated_data_multiply/",file_name)

    except Exception as e:
        print(e)
        return (-1,1)

# checks if f(x,y)=f(x+y)
def check_translational_symmetry_plus(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD \n")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+"/%s" %filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
                product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():
            min_error = 1000
            best_i = -1
            best_j = -1
            best_mu = 0
            best_sigma = 0
            for i in range(0,n_variables,1):
                for j in range(0,n_variables,1):
                    if i<j:
                        fact_translate = factors.clone()
                        a = 0.5*min(torch.std(fact_translate[:,i]),torch.std(fact_translate[:,j]))
                        fact_translate[:,i] = fact_translate[:,i] + a
                        fact_translate[:,j] = fact_translate[:,j] - a
                        list_errs = abs(product-model(fact_translate))
                        error = torch.median(list_errs)
                        mu = torch.mean(torch.log2(1+list_errs*2**30))
                        sigma = torch.std(torch.log2(1+list_errs*2**30))
                        if error<min_error:
                            min_error = error
                            best_i = i
                            best_j = j
                            best_mu = mu
                            best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma

    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)
    
def do_translational_symmetry_plus(pathdir, filename, i,j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+"/%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"/%s" %filename, usecols=(0,))

        for k in range(1,n_variables):
            v = np.loadtxt(pathdir+"/%s" %filename, usecols=(k,))
            variables = np.column_stack((variables,v))
        
        f_dependent = np.loadtxt(pathdir+"/%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        models_one = []
        models_rest = []

        with torch.no_grad():             
            file_name = filename + "-translated_plus"
            data_translated = variables
            ct_median =torch.median(torch.from_numpy(variables[:,j]))
            data_translated[:,i] = variables[:,i]+variables[:,j]
            data_translated =  np.delete(data_translated, j, axis=1)
            data_translated = np.column_stack((data_translated,f_dependent))
            try:
                os.mkdir("results/translated_data_plus/")
            except:
                pass
            np.savetxt("results/translated_data_plus/"+file_name , data_translated)
            remove_input_neuron(model,n_variables,j,ct_median,"results/NN_trained_models/models/"+filename + "-translated_plus_pretrained.h5")
            return ("results/translated_data_plus/", file_name)
        
    except Exception as e:
        print(e)
        return (-1,-1)



from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import pandas as pd
import numpy as np
import torch
from torch.utils import data
import pickle
from torch.optim.lr_scheduler import CosineAnnealingLR
from matplotlib import pyplot as plt
import time

is_cuda = torch.cuda.is_available()

bs = 2048

class MultDataset(data.Dataset):
    def __init__(self, factors, product):
        'Initialization'
        self.factors = factors
        self.product = product

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.product)

    def __getitem__(self, index):
        # Load data and get label
        x = self.factors[index]
        y = self.product[index]

        return x, y

def rmse_loss(pred, targ):
    denom = targ**2
    denom = torch.sqrt(denom.sum()/len(denom))

    return torch.sqrt(F.mse_loss(pred, targ))/denom


def NN_eval(pathdir,filename):
    try:
        n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+filename, usecols=(0,))

        if n_variables==0:
            return 0
        elif n_variables==1:
            variables = np.reshape(variables,(len(variables),1))
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+filename, usecols=(j,))
                variables = np.column_stack((variables,v))

        f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables[0:int(5*len(variables)/6)])
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()
        product = torch.from_numpy(f_dependent[0:int(5*len(f_dependent)/6)])
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        factors_val = torch.from_numpy(variables[int(5*len(variables)/6):int(len(variables))])
        if is_cuda:
            factors_val = factors_val.cuda()
        else:
            factors_val = factors_val
        factors_val = factors_val.float()
        product_val = torch.from_numpy(f_dependent[int(5*len(variables)/6):int(len(variables))])      
        if is_cuda:
            product_val = product_val.cuda()
        else:
            product_val = product_val
        product_val = product_val.float()

        class SimpleNet(nn.Module):
            def __init__(self, ni):
                super().__init__()
                self.linear1 = nn.Linear(ni, 128)
                self.linear2 = nn.Linear(128, 128)
                self.linear3 = nn.Linear(128, 64)
                self.linear4 = nn.Linear(64,64)
                self.linear5 = nn.Linear(64,1)

            def forward(self, x):
                x = F.tanh(self.linear1(x))
                x = F.tanh(self.linear2(x))
                x = F.tanh(self.linear3(x))
                x = F.tanh(self.linear4(x))
                x = self.linear5(x)
                return x

        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
                    
        model.load_state_dict(torch.load("results/NN_trained_models/models/"+filename+".h5"))
        model.eval()
        return(rmse_loss(model(factors_val),product_val),model)

    except Exception as e:
        print(e)
        return (100,0)








# Calculates the complexity of a number to be used for the Pareto frontier after snapping

import numpy as np
from .S_snap import bestApproximation

def get_number_DL_snapped(n):
    epsilon = 1e-10
    n = float(n)
    if np.isnan(n):
        return 1000000
    elif np.abs(n - int(n)) < epsilon:
        return np.log2(1 + abs(int(n)))
    elif np.abs(n - bestApproximation(n,10000)[0]) < epsilon:
        _, numerator, denominator, _ = bestApproximation(n, 10000)
        return np.log2((1 + abs(numerator)) * abs(denominator))
    elif np.abs(n - np.pi) < epsilon:
        return np.log2(1+3)
    else:
        PrecisionFloorLoss = 1e-14
        return np.log2(1 + (float(n) / PrecisionFloorLoss) ** 2) / 2






import numpy as np
from numpy import linalg, zeros, ones, hstack, asarray
import itertools
from matplotlib import pyplot as plt
from scipy.sparse.linalg import lsqr
import os
from sympy import symbols, Add, Mul, S


def basis_vector(n, i):
    x = zeros(n, dtype=int)
    x[i] = 1
    return x

def as_tall(x):
    return x.reshape(x.shape + (1,))


def multipolyfit(xs, y, deg):
    
    y = asarray(y).squeeze()
    rows = y.shape[0]
    xs = asarray(xs)
    try:
        num_covariates = xs.shape[1]
    except:
        num_covariates = 1
        xs = np.reshape(xs,(len(xs),1))

    xs = hstack((ones((xs.shape[0], 1), dtype=xs.dtype) , xs))
    
    generators = [basis_vector(num_covariates+1, i) for i in range(num_covariates+1)]
        
    # All combinations of degrees
    powers = map(sum, itertools.combinations_with_replacement(generators, deg))

    # Raise data to specified degree pattern, stack in order
    A = hstack(asarray([as_tall((xs**p).prod(1)) for p in powers]))
    params = lsqr(A, y)[0] # get the best params of the fit
    rms = lsqr(A, y)[4] # get the rms params of the fit
    
    return (params, rms)


def getBest(xs,y,max_deg):
    results = []
    for i in range(0,max_deg+1):
        results = results + [multipolyfit(xs,y,i)]
    results = np.array(results)
    # get the parameters and error of the fit with the lowest rms error
    params = results[np.argmin(results[:,1:])][0]
    error = results[np.argmin(results[:,1:])][1]
    deg = np.argmin(results[:,1:])
    return (params, error, deg)




import numpy as np
import os
from .S_run_bf_polyfit import run_bf_polyfit

def get_acos(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.arccos(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "acos")
    except:
        return PA

    return PA

def get_asin(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.arcsin(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "asin")
    except:
        return PA

    return PA

def get_atan(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.arctan(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "atan")
    except:
        return PA

    return PA


def get_cos(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.cos(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "cos")
    except:
        return PA

    return PA


def get_exp(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.exp(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "exp")
    except:
        return PA

    return PA


def get_inverse(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = 1/data[:,-1]
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "inverse")
    except:
        return PA

    return PA


def get_log(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.log(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "log")
    except:
        return PA

    return PA


def get_sin(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.sin(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "sin")
    except:
        return PA

    return PA


def get_sqrt(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.sqrt(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "sqrt")

    except:
        return PA

    return PA


def get_squared(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = data[:,-1]**2
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "squared")

    except:
        return PA

    return PA


def get_tan(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3):
    try:
        os.mkdir(pathdir_write_to)
    except:
        pass
    data = np.loadtxt(pathdir+filename)
    try:
        data[:,-1] = np.tan(data[:,-1])
        np.savetxt(pathdir_write_to+filename,data)
        PA = run_bf_polyfit(pathdir,pathdir_write_to,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg, "tan")

    except:
        return PA

    return PA






import os
import re
import shutil
import urllib.request


def get_demos(dataset="example_data"):
    base = "https://space.mit.edu/home/tegmark/aifeynman/" + dataset
    with urllib.request.urlopen(base) as base_response:
        string = base_response.read().decode('utf-8')

        # the pattern actually creates duplicates in the list
        pattern = re.compile('>.*\.txt<')
        filelist = map(lambda x: x[1:-1], pattern.findall(string))

        print("downloading...")

        try:
            os.mkdir(dataset)
        except FileExistsError:
            pass

        for fname in filelist:
            print(fname)
            print(base + '/' + fname)
            print(dataset + '/' + fname)
            with urllib.request.urlopen(base + '/' + fname) as response, \
                    open(dataset + '/' + fname, 'wb') as out_file:
                shutil.copyfileobj(response, out_file)


if __name__ == "__main__":
    get_demos()



# SAve a file with 2*(n-1) columns contaning the (n-1) independent variables and the (n-1) gradients of the trained NN with respect these variables

import matplotlib.pyplot as plt
import numpy as np
import copy
import os
import sys
import torch
import torch.nn as nn
import torch.nn.functional as F
is_cuda = torch.cuda.is_available()

def evaluate_derivatives(pathdir,filename,model):
    try:
        data = np.loadtxt(pathdir+filename)[:,0:-1]
        pts = np.loadtxt(pathdir+filename)[:,0:-1]
        pts = torch.tensor(pts)
        pts = pts.clone().detach()
        is_cuda = torch.cuda.is_available()
        grad_weights = torch.ones(pts.shape[0], 1)
        if is_cuda:
            pts = pts.float().cuda()
            model = model.cuda()
            grad_weights = grad_weights.cuda()

        pts.requires_grad_(True)
        outs = model(pts)
        grad = torch.autograd.grad(outs, pts, grad_outputs=grad_weights, create_graph=True)[0]
        save_grads = grad.detach().data.cpu().numpy()
        save_data = np.column_stack((data,save_grads))
        np.savetxt("results/gradients_comp_%s.txt" %filename,save_data)
        return 1
    except:
        return 0



# add a function to compte complexity

from .get_pareto import Point, ParetoSet
from .RPN_to_pytorch import RPN_to_pytorch
from .RPN_to_eq import RPN_to_eq
import numpy as np
import matplotlib.pyplot as plt
from .S_brute_force import brute_force
from .S_get_number_DL_snapped import get_number_DL_snapped
from sympy.parsing.sympy_parser import parse_expr
from sympy import preorder_traversal, count_ops
from .S_polyfit import polyfit
from .S_get_symbolic_expr_error import get_symbolic_expr_error
from .S_add_sym_on_pareto import add_sym_on_pareto
from .S_add_snap_expr_on_pareto import add_snap_expr_on_pareto
import os
from os import path


def run_bf_polyfit(pathdir,pathdir_transformed,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg=3, output_type=""):
    input_data = np.loadtxt(pathdir_transformed+filename)
#############################################################################################################################
    if np.isnan(input_data).any()==False:
        # run BF on the data (+)
        print("Checking for brute force + \n")
        brute_force(pathdir_transformed,filename,BF_try_time,BF_ops_file_type,"+")

        try:
            # load the BF output data
            bf_all_output = np.loadtxt("results.dat", dtype="str")
            express = bf_all_output[:,2]
            prefactors = bf_all_output[:,1]
            prefactors = [str(i) for i in prefactors]

            # Calculate the complexity of the bf expression the same way as for gradient descent case
            complexity = []
            errors = []
            eqns = []
            for i in range(len(prefactors)):
                try:
                    if output_type=="":
                        eqn = prefactors[i] + "+" + RPN_to_eq(express[i])
                    elif output_type=="acos":
                        eqn = "cos(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="asin":
                        eqn = "sin(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="atan":
                        eqn = "tan(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="cos":
                        eqn = "acos(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="exp":
                        eqn = "log(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="inverse":
                        eqn = "1/(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="log":
                        eqn = "exp(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="sin":
                        eqn = "asin(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="sqrt":
                        eqn = "(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")**2"
                    elif output_type=="squared":
                        eqn = "sqrt(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="tan":
                        eqn = "atan(" + prefactors[i] + "+" + RPN_to_eq(express[i]) + ")"

                    eqns = eqns + [eqn]
                    errors = errors + [get_symbolic_expr_error(input_data,eqn)]
                    expr = parse_expr(eqn)
                    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
                    numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]
                    compl = 0
                    for j in numbers_expr:
                        try:
                            compl = compl + get_number_DL_snapped(float(j))
                        except:
                            compl = compl + 1000000

                    # Add the complexity due to symbols
                    n_variables = len(expr.free_symbols)
                    n_operations = len(count_ops(expr,visual=True).free_symbols)
                    if n_operations!=0 or n_variables!=0:
                        compl = compl + (n_variables+n_operations)*np.log2((n_variables+n_operations))

                    complexity = complexity + [compl]
                except:
                    continue

            for i in range(len(complexity)):
                PA.add(Point(x=complexity[i], y=errors[i], data=eqns[i]))

            # run gradient descent of BF output parameters and add the results to the Pareto plot
            for i in range(len(express)):
                try:
                    bf_gd_update = RPN_to_pytorch(input_data,eqns[i])
                    PA.add(Point(x=bf_gd_update[1],y=bf_gd_update[0],data=bf_gd_update[2]))
                except:
                    continue
        except:
            pass

    #############################################################################################################################
        # run BF on the data (*)
        print("Checking for brute force * \n")
        brute_force(pathdir_transformed,filename,BF_try_time,BF_ops_file_type,"*")

        try:
            # load the BF output data
            bf_all_output = np.loadtxt("results.dat", dtype="str")
            express = bf_all_output[:,2]
            prefactors = bf_all_output[:,1]
            prefactors = [str(i) for i in prefactors]

            # Calculate the complexity of the bf expression the same way as for gradient descent case
            complexity = []
            errors = []
            eqns = []
            for i in range(len(prefactors)):
                try:
                    if output_type=="":
                        eqn = prefactors[i] + "*" + RPN_to_eq(express[i])
                    elif output_type=="acos":
                        eqn = "cos(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="asin":
                        eqn = "sin(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="atan":
                        eqn = "tan(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="cos":
                        eqn = "acos(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="exp":
                        eqn = "log(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="inverse":
                        eqn = "1/(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="log":
                        eqn = "exp(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="sin":
                        eqn = "asin(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="sqrt":
                        eqn = "(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")**2"
                    elif output_type=="squared":
                        eqn = "sqrt(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"
                    elif output_type=="tan":
                        eqn = "atan(" + prefactors[i] + "*" + RPN_to_eq(express[i]) + ")"

                    eqns = eqns + [eqn]
                    errors = errors + [get_symbolic_expr_error(input_data,eqn)]
                    expr = parse_expr(eqn)
                    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
                    numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]
                    compl = 0
                    for j in numbers_expr:
                        try:
                            compl = compl + get_number_DL_snapped(float(j))
                        except:
                            compl = compl + 1000000

                    # Add the complexity due to symbols
                    n_variables = len(expr.free_symbols)
                    n_operations = len(count_ops(expr,visual=True).free_symbols)
                    if n_operations!=0 or n_variables!=0:
                        compl = compl + (n_variables+n_operations)*np.log2((n_variables+n_operations))

                    complexity = complexity + [compl]
                except:
                    continue

            # add the BF output to the Pareto plot
            for i in range(len(complexity)):
                PA.add(Point(x=complexity[i], y=errors[i], data=eqns[i]))

            # run gradient descent of BF output parameters and add the results to the Pareto plot
            for i in range(len(express)):
                try:
                    bf_gd_update = RPN_to_pytorch(input_data,eqns[i])
                    PA.add(Point(x=bf_gd_update[1],y=bf_gd_update[0],data=bf_gd_update[2]))
                except:
                    continue
        except:
            pass

    #############################################################################################################################
        # run polyfit on the data
        print("Checking polyfit \n")
        try:
            polyfit_result = polyfit(polyfit_deg, pathdir_transformed+filename)
            eqn = str(polyfit_result[0])

            # Calculate the complexity of the polyfit expression the same way as for gradient descent case
            if output_type=="":
                eqn = eqn
            elif output_type=="acos":
                eqn = "cos(" + eqn + ")"
            elif output_type=="asin":
                eqn = "sin(" + eqn + ")"
            elif output_type=="atan":
                eqn = "tan(" + eqn + ")"
            elif output_type=="cos":
                eqn = "acos(" + eqn + ")"
            elif output_type=="exp":
                eqn = "log(" + eqn + ")"
            elif output_type=="inverse":
                eqn = "1/(" + eqn + ")"
            elif output_type=="log":
                eqn = "exp(" + eqn + ")"
            elif output_type=="sin":
                eqn = "asin(" + eqn + ")"
            elif output_type=="sqrt":
                eqn = "(" + eqn + ")**2"
            elif output_type=="squared":
                eqn = "sqrt(" + eqn + ")"
            elif output_type=="tan":
                eqn = "atan(" + eqn + ")"

            polyfit_err = get_symbolic_expr_error(input_data,eqn)
            expr = parse_expr(eqn)
            is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
            numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]
            complexity = 0
            for j in numbers_expr:
                complexity = complexity + get_number_DL_snapped(float(j))
            try:
                # Add the complexity due to symbols
                n_variables = len(polyfit_result[0].free_symbols)
                n_operations = len(count_ops(polyfit_result[0],visual=True).free_symbols)
                if n_operations!=0 or n_variables!=0:
                    complexity = complexity + (n_variables+n_operations)*np.log2((n_variables+n_operations))
            except:
                pass

            #run zero snap on polyfit output
            PA_poly = ParetoSet()
            PA_poly.add(Point(x=complexity, y=polyfit_err, data=str(eqn)))
            PA_poly = add_snap_expr_on_pareto(pathdir, filename, str(eqn), PA_poly)

            for l in range(len(PA_poly.get_pareto_points())):
                PA.add(Point(PA_poly.get_pareto_points()[l][0],PA_poly.get_pareto_points()[l][1],PA_poly.get_pareto_points()[l][2]))

        except:
            pass

        print("Complexity  RMSE  Expression")
        for pareto_i in range(len(PA.get_pareto_points())):
            print(PA.get_pareto_points()[pareto_i])

        return PA
    else:
        return PA



import matplotlib.pyplot as plt
import numpy as np
import copy
import os
import sys
import torch
import torch.nn as nn

from itertools import chain, combinations, islice
from scipy.stats import mannwhitneyu
from sklearn.neighbors import KernelDensity
from scipy.stats import iqr


from collections import Counter, namedtuple, OrderedDict
from copy import deepcopy

import torch.utils.data as data_utils


'''
This part discovers symmetry using the model fitted
on data
'''

def powerset_atleast_2(iterable, max_subset_size):
    "powerset([1,2,3]) -->  (1,2) (1,3) (2,3)"
    s = list(iterable)
    r = chain.from_iterable(combinations(s, r) for r in range(2, max_subset_size+1))
    # return islice(r, 0, 3)
    return r

def evaluate_derivatives(model, s,  pts):
    
    pts = pts.clone().detach()
    try:
        device = 'cuda' if model.is_cuda else 'cpu'
    except:
        device = 'cpu'
    pts = pts.to(device=device)

    pts.requires_grad_(True)
    outs = torch.zeros(pts.shape, device=device)
    if pts.grad is not None:
        pts.grad.data.zero_()
    for i in range(pts.shape[0]):
        model.eval()
        d = pts[[i], :]
        d.requires_grad_(True)
        r = model(d)
        r.backward()
    return pts.grad[:, s].clone().detach()

def evaluate_derivatives_andrew(model, s, pts):
    pts = pts.clone().detach()
    is_cuda = torch.cuda.is_available()
    grad_weights = torch.ones(pts.shape[0], 1)
    if is_cuda:
        pts = pts.cuda()
        model = model.cuda()
        grad_weights = grad_weights.cuda()
    
    pts.requires_grad_(True)
    outs = model(pts)
    grad = torch.autograd.grad(outs, pts, grad_outputs=grad_weights, create_graph=True)[0]
    return grad[:, s].detach()

def build_true_model(func):
    class TrueModel(nn.Module):
        def forward(self, X):
            return func(*[X[:,[i]] for i in range(X.shape[1])])
    return TrueModel()


def draw_samples(X, y, model, s, NUM_SAMPLES, point = None):
    '''
    Draw samples by sampling each dimension independently, 
    keeping the positions at s fixed to given point if exists,
    sampled point if not.
    '''
    n = X.shape[1]
    is_cuda = X.is_cuda
    device = 'cuda' if is_cuda else 'cpu'
    if point is None:
        idx = torch.randint(0, X.shape[0], (1,)).repeat(len(s))
    else:
        idx = torch.tensor([point]).repeat(len(s))
    pts = torch.randint(0, X.shape[0], (NUM_SAMPLES, X.shape[1]))
    pts[:, s] = idx
    actual_pts = torch.zeros(NUM_SAMPLES, X.shape[1], dtype=torch.float)
    for i in range(NUM_SAMPLES):
        for j in range(X.shape[1]):
            actual_pts[i,j] = X[pts[i,j],j]
    return actual_pts.to(device=device)

def visualize_score(normalized_grads, overall_score):
    fig, ax = plt.subplots(figsize=(6, 3), subplot_kw=dict(aspect="equal"))
    ax.scatter([t[0] for t in normalized_grads], [t[1] for t in normalized_grads])
    ax.set_title(f'Score: {1-overall_score}')
    ax.set_xlim((-1, 1))
    ax.set_ylim((-1, 1))
    plt.show()

def score_consistency(grads_tensor):
    n_pts = grads_tensor.shape[0]
    grads_tensor = grads_tensor.cpu()
    norms = [np.linalg.norm(grads_tensor[i, :].numpy()) for i in range(n_pts)]
    normalized_grads = [grads_tensor[i,:].numpy()/norms[i] for i in range(n_pts)]
    A = np.array(normalized_grads)
    D = np.einsum('ij,ik', A, A)
    evals, evecs = np.linalg.eig(D)
    nv = evals.shape[0]
    assert(nv == A.shape[1])
    # evecs should be d x (num eigenvectors - probably also d)
    dots = np.einsum('ij,jk->ik', A, evecs)
    scores = np.einsum('ij,ij->j', dots, dots)
    overall_score = np.max(scores)/n_pts
    mean_vec = evecs[:,np.argmax(scores)].flatten()
    return overall_score, mean_vec


plot_counter = 0
def visualize_distribution(hypot, bench):
    global plot_counter
    eps=1e-10
    hypot=np.array(hypot)
    bench = np.array(bench)
    kde = KernelDensity(kernel='tophat', bandwidth=0.25).fit(np.log10(1-hypot+eps)[:,np.newaxis])
    kde2 = KernelDensity(kernel='tophat', bandwidth=0.25).fit(np.log10(1-bench+eps)[:,np.newaxis])
    X_plot = np.linspace(-8, 0, 1000)[:, np.newaxis]
    log_dens = kde.score_samples(X_plot)
    log_dens2= kde2.score_samples(X_plot)
    plt.plot(X_plot[:, 0], np.exp(log_dens), color='#FFAA00', label='Hypothesis') #hypot is orange
    plt.plot(X_plot[:, 0], np.exp(log_dens2), color='#00AAFF', label='Benchmark')#blue is bench
    plt.legend()
    plt.savefig(f'plot_{plot_counter}.pdf')
    plot_counter += 1
    plt.show()

def get_kde(X_vals, dist, kernel='tophat', bandwidth=0.3):
    eps=1e-15
    scores = np.array(dist)
    kde = KernelDensity(kernel=kernel, bandwidth=bandwidth).fit(np.log10(1-scores+eps)[:,np.newaxis])
    log_dens = kde.score_samples(X_vals)
    return np.exp(log_dens)

def signal_to_noise(high, low):
    '''
    Parameters: list of scores for hypothesis, list of scores for benchmark
    Returns a non-negative value
    '''
    eps = 1e-10
    def transform_scores(scores):
        scores = np.array(scores)
        scores = np.log10(1-scores + eps)
        return scores
    high = transform_scores(high)
    low = transform_scores(low)
    low_tail  = iqr(low, rng=(10,50))
    low_med  = np.median(low)
    high_med  = np.median(high)
    # print(high_med, low_med, low_tail)

    if high_med >= low_med:
        return 0
    else:
        # return (low_med - high_med) / low_tail
        return 0 - high_med

score_distributions = {}

def filter_decompositions_relative_scoring(X, y, model, max_subset_size=None, visualize=False):
    '''
    X: torch tensor. N * d
    y: torch tensor. N * 1
    model: torch nn.module
    returns: [(score, subset)],
    where subset is a tuple of indices in ascending order
    '''
    is_cuda = torch.cuda.is_available()
    if is_cuda:
        X = X.cuda()
        y = y.cuda()
        model = model.cuda()
        # print('Using cuda')

    n = X.shape[1]
    NUM_TRIALS = 200
    NUM_SAMPLES =  60
    results = []
    max_subset_size = n-1 if max_subset_size is None else max_subset_size
    random_indices = np.random.choice(X.shape[0], size=min(NUM_TRIALS, X.shape[0]), replace=False)
    for s in powerset_atleast_2(range(0, n), max_subset_size=max_subset_size):
        print(f'Trying {s}')
        inv_s = tuple([i for i in range(0, n) if i not in s])
        hypot_scores = []
        bench_scores = []
        for i in range(random_indices.shape[0]):
            samples = draw_samples(X, y, model, s, NUM_SAMPLES, point=random_indices[i])
            score, _ = score_consistency(evaluate_derivatives_andrew(model, s, samples))
            hypot_scores.append(score)
        for i in range(random_indices.shape[0]):
            samples = draw_samples(X, y, model, (), NUM_SAMPLES, point=random_indices[i])
            score, _ = score_consistency(evaluate_derivatives_andrew(model, s, samples))
            bench_scores.append(score)
        snr = signal_to_noise(hypot_scores, bench_scores)
        # penalizes larger decompositions
        snr -= np.log10(2)*len(s) 
        results.append((snr, s))
        print((snr, s))
        if visualize:
            print(snr)
            visualize_distribution(hypot_scores, bench_scores)
            score_distributions[s] = (snr, hypot_scores)
    results.sort(key=lambda x:-x[0])
    return results

'''
Some helper functions
'''

def invert_subset(s, n):
    return tuple([i for i in range(n) if i not in s])
def to_bin_str(s, n):
    return ''.join(['1' if i in s else '0' for i in range(n)])
def to_numpy_mask(s, n):
    return np.array([i in s for i in range(n)])


def extract_gradients(X, y, model, s, num_points):
    is_cuda = torch.cuda.is_available()
    if is_cuda:
        X = X.cuda()
        y = y.cuda()
        model = model.cuda()
    idx = np.random.randint(0, X.shape[0], size=num_points)
    gradients = np.zeros((num_points, len(s)))
    for i in range(num_points):
        samples = draw_samples(X, y, model, s, 50, point=idx[i])
        _, gradients[i, :] = score_consistency(evaluate_derivatives_andrew(model, s, samples))
        # normalize first dimension
        if(gradients[i, 0] < 0):
             gradients[i,:] *= -1
    return np.concatenate((X[idx, :][:, s].cpu().data.numpy(), gradients),axis=1)


'''
Actual hook
Returns pair (numpy with 2k columns. First k are data points, next k are gradients), (bitmask as numpy array)
'''

def identify_decompositions(pathdir,filename, model, max_subset_size=2, visualize=False):
    print("identify_decompositions",pathdir,filename)
    data = np.loadtxt(pathdir+filename)
    X = torch.Tensor(data[:, :-1])
    y = torch.Tensor(data[:, [-1]])
    # Return best decomposition                                                                                                                                                   
    all_scores = filter_decompositions_relative_scoring(X, y, model, visualize=visualize)
    assert(all_scores)
    best_decomposition = all_scores[0][1]
    gradients = extract_gradients(X, y, model, best_decomposition, 10000)
    np.savetxt("results/gradients_gen_sym_%s" %filename, gradients)
    ll = np.arange(0,X.shape[1],1)
    print("mask", to_numpy_mask(best_decomposition, X.shape[1]))
    return ll[to_numpy_mask(best_decomposition, X.shape[1])]




from __future__ import print_function
import torch
import os
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import pandas as pd
import numpy as np
import torch
from torch.utils import data
import pickle
from torch.optim.lr_scheduler import CosineAnnealingLR
from matplotlib import pyplot as plt
from itertools import combinations
import time

is_cuda = torch.cuda.is_available()

class SimpleNet(nn.Module):
    def __init__(self, ni):
        super().__init__()
        self.linear1 = nn.Linear(ni, 128)
        self.linear2 = nn.Linear(128, 128)
        self.linear3 = nn.Linear(128, 64)
        self.linear4 = nn.Linear(64,64)
        self.linear5 = nn.Linear(64,1)

    def forward(self, x):
        x = F.tanh(self.linear1(x))
        x = F.tanh(self.linear2(x))
        x = F.tanh(self.linear3(x))
        x = F.tanh(self.linear4(x))
        x = self.linear5(x)
        return x

def rmse_loss(pred, targ):
    denom = targ**2
    denom = torch.sqrt(denom.sum()/len(denom))
    return torch.sqrt(F.mse_loss(pred, targ))/denom

def check_separability_plus(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        # make some variables at the time equal to the median of factors
        models_one = []
        models_rest = []

        with torch.no_grad():
            fact_vary = factors.clone()
            for k in range(len(factors[0])):
                fact_vary[:,k] = torch.full((len(factors),),torch.median(factors[:,k]))

            # loop through all indices combinations
            var_indices_list = np.arange(0,n_variables,1)
            min_error = 1000
            best_i = []
            best_j = []
            best_mu = 0
            best_sigma = 0
            for i in range(1,n_variables):
                c = combinations(var_indices_list, i)
                for j in c:
                    fact_vary_one = factors.clone()
                    fact_vary_rest = factors.clone()
                    rest_indx = list(filter(lambda x: x not in j, var_indices_list))
                    for t1 in rest_indx:
                        fact_vary_one[:,t1] = torch.full((len(factors),),torch.median(factors[:,t1]))
                    for t2 in j:
                        fact_vary_rest[:,t2] = torch.full((len(factors),),torch.median(factors[:,t2]))
                    # check if the equation is separable
                    sm = model(fact_vary_one)+model(fact_vary_rest)
                    #error = torch.sqrt(torch.mean((product-sm+model(fact_vary))**2))/torch.sqrt(torch.mean(product**2))
                    list_errs = 2*abs(product-sm+model(fact_vary))
                    error = torch.median(list_errs)
                    mu = torch.mean(torch.log2(1+list_errs*2**30))
                    sigma = torch.std(torch.log2(1+list_errs*2**30))
                    #error = 2*torch.median(abs(product-sm+model(fact_vary)))
                    if error<min_error:
                        min_error = error
                        best_i = j
                        best_j = rest_indx
                        best_mu = mu
                        best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma
                        
    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)                    
                    
                                           
def do_separability_plus(pathdir, filename, list_i,list_j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables) 
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        # make some variables at the time equal to the median of factors
        models_one = []
        models_rest = []
        
        fact_vary = factors.clone()
        for k in range(len(factors[0])):
            fact_vary[:,k] = torch.full((len(factors),),torch.median(factors[:,k]))
        fact_vary_one = factors.clone()
        fact_vary_rest = factors.clone()
        for t1 in list_j:
            fact_vary_one[:,t1] = torch.full((len(factors),),torch.median(factors[:,t1]))
        for t2 in list_i:
            fact_vary_rest[:,t2] = torch.full((len(factors),),torch.median(factors[:,t2]))

        with torch.no_grad():
            str1 = filename+"-add_a"
            str2 = filename+"-add_b"
            # save the first half
            data_sep_1 = variables
            data_sep_1 = np.delete(data_sep_1,list_j,axis=1)
            data_sep_1 = np.column_stack((data_sep_1,model(fact_vary_one).cpu()))
            # save the second half  
            data_sep_2 = variables
            data_sep_2 = np.delete(data_sep_2,list_i,axis=1)
            data_sep_2 = np.column_stack((data_sep_2,model(fact_vary_rest).cpu()-model(fact_vary).cpu()))
            try:
                os.mkdir("results/separable_add/")
            except:
                pass
            np.savetxt("results/separable_add/"+str1,data_sep_1)
            np.savetxt("results/separable_add/"+str2,data_sep_2)
            # if it is separable, return the 2 new files created and the index of the column with the separable variable
            return ("results/separable_add/",str1,"results/separable_add/",str2)

    except Exception as e:
        print(e)
        return (-1,-1)

        
def check_separability_multiply(pathdir, filename):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))

        # Pick only data which is close enough to the maximum value (5 times less or higher)                                                                   
        max_output = np.max(abs(f_dependent))
        use_idx = np.where(abs(f_dependent)>=max_output/5)
        f_dependent = f_dependent[use_idx]
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))
        variables = variables[use_idx]

        factors = torch.from_numpy(variables)
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        # make some variables at the time equal to the median of factors
        models_one = []
        models_rest = []

        with torch.no_grad():
            fact_vary = factors.clone()
            for k in range(len(factors[0])):
                fact_vary[:,k] = torch.full((len(factors),),torch.median(factors[:,k]))

            # loop through all indices combinations
            var_indices_list = np.arange(0,n_variables,1)
            min_error = 1000
            best_i = []
            best_j = []
            best_mu = 0
            best_sigma = 0
            for i in range(1,n_variables):
                c = combinations(var_indices_list, i)
                for j in c:
                    fact_vary_one = factors.clone()
                    fact_vary_rest = factors.clone()
                    rest_indx = list(filter(lambda x: x not in j, var_indices_list))
                    for t1 in rest_indx:
                        fact_vary_one[:,t1] = torch.full((len(factors),),torch.median(factors[:,t1]))
                    for t2 in j:
                        fact_vary_rest[:,t2] = torch.full((len(factors),),torch.median(factors[:,t2]))
                    # check if the equation is separable
                    pd = model(fact_vary_one)*model(fact_vary_rest)
                    #error = torch.sqrt(torch.mean((product-pd/model(fact_vary))**2))/torch.sqrt(torch.mean(product**2))
                    list_errs = 2*abs(product-pd/model(fact_vary))
                    error = torch.median(list_errs)
                    mu = torch.mean(torch.log2(1+list_errs*2**30))
                    sigma = torch.std(torch.log2(1+list_errs*2**30))
                    if error<min_error:
                        min_error = error
                        best_i = j
                        best_j = rest_indx
                        best_mu = mu
                        best_sigma = sigma
        return min_error, best_i, best_j, best_mu, best_sigma
                    
    except Exception as e:
        print(e)
        return (-1,-1,-1,-1,-1)                         

                    
                    
def do_separability_multiply(pathdir, filename, list_i,list_j):
    try:
        pathdir_weights = "results/NN_trained_models/models/"

        # load the data
        n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+filename, usecols=(0,))

        if n_variables==1:
            print(filename, "just one variable for ADD")
            # if there is just one variable you have nothing to separate
            return (-1,-1,-1)
        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+filename, usecols=(j,))
                variables = np.column_stack((variables,v))
        

        f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables)
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        # load the trained model and put it in evaluation mode
        if is_cuda:
            model = SimpleNet(n_variables).cuda()
        else:
            model = SimpleNet(n_variables)
        model.load_state_dict(torch.load(pathdir_weights+filename+".h5"))
        model.eval()

        # make some variables at the time equal to the median of factors
        models_one = []
        models_rest = []                    

        fact_vary = factors.clone()
        for k in range(len(factors[0])):
            fact_vary[:,k] = torch.full((len(factors),),torch.median(factors[:,k]))
        fact_vary_one = factors.clone()
        fact_vary_rest = factors.clone()
        for t1 in list_j:
            fact_vary_one[:,t1] = torch.full((len(factors),),torch.median(factors[:,t1]))
        for t2 in list_i:
            fact_vary_rest[:,t2] = torch.full((len(factors),),torch.median(factors[:,t2]))        
        
        with torch.no_grad():
            str1 = filename+"-mult_a"
            str2 = filename+"-mult_b"
            # save the first half
            data_sep_1 = variables
            data_sep_1 = np.delete(data_sep_1,list_j,axis=1)
            data_sep_1 = np.column_stack((data_sep_1,model(fact_vary_one).cpu()))
            # save the second half  
            data_sep_2 = variables
            data_sep_2 = np.delete(data_sep_2,list_i,axis=1)
            data_sep_2 = np.column_stack((data_sep_2,model(fact_vary_rest).cpu()/model(fact_vary).cpu()))
            try:
                os.mkdir("results/separable_mult/")
            except:
                pass
            np.savetxt("results/separable_mult/"+str1,data_sep_1)
            np.savetxt("results/separable_mult/"+str2,data_sep_2)
            # if it is separable, return the 2 new files created and the index of the column with the separable variable
            return ("results/separable_mult/",str1,"results/separable_mult/",str2)

    except Exception as e:
        print(e)
        return (-1,-1)

        



# Calculates the complexity of a number to be used for the Pareto frontier

import numpy as np 

def get_number_DL(n):
    epsilon = 1e-10
    # check if integer
    if np.isnan(n):
        return 1000000
    elif np.abs(n - int(n)) < epsilon:
        return np.log2(1+abs(n))
    elif np.abs(n - np.pi) < epsilon:
        return np.log2(1+3)
    # check if real
    else:
        PrecisionFloorLoss = 1e-14
        return np.log2(1 + (float(n) / PrecisionFloorLoss) ** 2) / 2




import numpy as np
from .RPN_to_eq import RPN_to_eq
from scipy.optimize import fsolve
from sympy import lambdify, N
import torch
import copy
import torch.nn as nn
import torch.nn.functional as F
from .get_pareto import Point, ParetoSet
from .S_get_expr_complexity import get_expr_complexity
from . import test_points
import os
import warnings
warnings.filterwarnings("ignore")
is_cuda = torch.cuda.is_available()


# fix this to work with the other variables constant
def check_gen_sym(pathdir,filename,model,gen_sym_idx,express,mu,sigma,nu=10):
    gen_sym_idx = np.append(gen_sym_idx,-1)
    data_all = np.loadtxt(pathdir+filename)
    # Choose only the data to be separated
    data = np.loadtxt(pathdir+filename)[:,gen_sym_idx]
    # Turn the equation from RPN to normal mathematical expression
    eq = RPN_to_eq(express)
    print(eq)
    # Get the variables appearing in the equation
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    variables = []
    N_vars = len(data[0])-1
    for i in range(N_vars):
        variables = variables + [possible_vars[i]]
    symbols = variables
    f = lambdify(symbols, N(eq))

    fixed = data[:,0:-1]
    length_fixed = len(fixed)

    bm = np.ones(len(data[0])-1,dtype=bool)
    obj = test_points.init_general_test_point(eq, data[:,:-1], data[:,-1], bm)

    list_z = np.array([])
    z = 0
    i = 0
    while z<nu and i<len(data[0:1000]):
        # Generate functions based on the discovered possible equation and check if they are right
        dt = test_points.get_test_point(obj,data[i][:-1])
        diff = abs(f(*fixed[i])-f(*dt))
        with torch.no_grad():
            if diff<1e-4:
                if is_cuda:
                    dt_ = data_all[i]
                    ii = 0
                    for k in gen_sym_idx[:-1]:
                        dt_[k]=dt[ii]
                        ii = ii + 1
                    dt = torch.tensor(dt_).float().cuda().view(1,len(dt_))
                    dt = torch.cat((torch.tensor([np.zeros(len(dt[0]))]).float().cuda(),dt), 0)
                    error = torch.tensor(data[:,-1][i]).cuda()-model(dt[:,:-1])[1:]
                    error = error.cpu().detach().numpy()
                    list_z = np.append(list_z,np.log2(1+abs(error)*2**30))
                    z = np.sqrt(len(list_z))*(np.mean(list_z)-mu)/sigma
                else:
                    dt_ = data_all[i]
                    ii = 0
                    for k in gen_sym_idx[:-1]:
                        dt_[k]=dt[ii]
                        ii = ii + 1
                    dt = torch.tensor(dt_).float().view(1,len(dt_))
                    dt = torch.cat((torch.tensor([np.zeros(len(dt[0]))]).float(),dt), 0)
                    error =torch.tensor(data[:,-1][i])-model(dt[:,:-1])[1:]
                    error = error.detach().numpy()
                    list_z = np.append(list_z,np.log2(1+abs(error)*2**30))
                    z = np.sqrt(len(list_z))*(np.mean(list_z)-mu)/sigma
                    print(list_z)
                    print(z)
                    print("")
                i = i + 1
            else:
                i = i + 1

    print(i)

    if i==len(data[0:1000]) and np.mean(list_z)<mu:
        return (1,express,np.mean(list_z),np.std(list_z))
    else:
        return (0,express,100,100)


#express = "bba*>/"
#print(check_gen_sym("../","comp_data_2.txt_train",model,[0,1],express,100,1))#[0,1,-1]))


def do_gen_sym(pathdir, filename, gen_sym_idx,express):
    gen_sym_idx = np.append(gen_sym_idx,-1)
    data_all = np.loadtxt(pathdir+filename)

    # Choose only the data to be separated
    data = np.loadtxt(pathdir+filename)[:,gen_sym_idx]
    # Turn the equation from RPN to normal mathematical expression
    eq = RPN_to_eq(express)
    # Get the variables appearing in the equation
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    variables = []

    N_vars = len(data[0])-1
    for i in range(N_vars):
        variables = variables + [possible_vars[i]]

    symbols = variables
    f = lambdify(symbols, N(eq))

    ii = 0
    for k in gen_sym_idx[1:-1]:
        data_all = np.delete(data_all,k-ii,1)
        ii = ii + 1

    new_data = f(*np.transpose(data[:,0:-1]))
    data_all[:,gen_sym_idx[0]]=new_data
    #save_data = np.column_stack((new_data,data_all))
    save_data = data_all

    try:
        os.mkdir("results/gen_sym")
    except:
        pass

    file_name = filename + "-gen_sym"
    np.savetxt("results/gen_sym/"+file_name,save_data)

    return ("results/gen_sym/", file_name)

#print(do_gen_sym("../","comp_data_2.txt_train",[1,2,-1]))#[ True, True, False]))


def add_gen_sym_on_pareto(PA1,PA, gen_sym_idx, express):
    # Turn the equation from RPN to normal mathematical expression
    possible_vars = ["x%s" %i for i in np.arange(0,100,1)]
    gen_sym_idx = np.array(gen_sym_idx)
    math_eq = RPN_to_eq(express)

    PA1 = np.array(PA1.get_pareto_points()).astype('str')
    for i in range(len(PA1)):
        exp1 = PA1[i][2]
        temp_list = copy.deepcopy(gen_sym_idx)
        bf_eq = math_eq
        print(exp1, bf_eq,gen_sym_idx)
        while(len(temp_list)>1):
            for j in range(len(possible_vars)-len(temp_list),temp_list[-1]-len(temp_list)+1,-1):
                exp1 = exp1.replace(possible_vars[j],possible_vars[j+1])
            temp_list = np.delete(temp_list,-1)
            print("check: ",exp1, temp_list)
        # replace variables in bf_eq
        arr_idx = np.flip(np.arange(0,len(gen_sym_idx),1), axis=0)
        actual_idx = np.flip(gen_sym_idx, axis=0)
        for k in range(len(gen_sym_idx)):
            bf_eq = bf_eq.replace(possible_vars[arr_idx[k]],possible_vars[actual_idx[k]])

        exp1 = exp1.replace(possible_vars[temp_list[0]],"(" + bf_eq + ")")
        compl = get_expr_complexity(exp1)
        PA.add(Point(x=compl,y=float(PA1[i][1]),data=str(exp1)))

    return PA



# Combines 2 pareto fromtier obtained from the separability test into a new one.

from .get_pareto import Point, ParetoSet
from .S_get_symbolic_expr_error import get_symbolic_expr_error
from sympy.parsing.sympy_parser import parse_expr
import numpy as np
import matplotlib.pyplot as plt
import os
from os import path
from sympy import Symbol, lambdify, N
from .get_pareto import Point, ParetoSet
from .S_get_expr_complexity import get_expr_complexity

def combine_pareto(input_data,PA1,PA2,idx_list_1,idx_list_2,PA,sep_type = "+"):
    possible_vars = ["x%s" %i for i in np.arange(0,30,1)]
    PA1 = np.array(PA1.get_pareto_points()).astype('str')
    PA2 = np.array(PA2.get_pareto_points()).astype('str')
    for i in range(len(PA1)):
        for j in range(len(PA2)):
            try:
                # replace the variables from the separated parts with the variables reflecting the new combined equation
                exp1 = PA1[i][2]
                exp2 = PA2[j][2]
                for k in range(len(idx_list_1)-1,-1,-1):
                    exp1 = exp1.replace(possible_vars[k],possible_vars[idx_list_1[k]])
                for k in range(len(idx_list_2)-1,-1,-1):
                    exp2 = exp2.replace(possible_vars[k],possible_vars[idx_list_2[k]])
                new_eq = "(" + exp1 + ")" + sep_type + "(" + exp2 + ")"
                compl = get_expr_complexity(new_eq)
                PA.add(Point(x=compl,y=get_symbolic_expr_error(input_data,new_eq),data=new_eq))
            except:
                continue
    return PA







from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import pandas as pd
import numpy as np
import torch
from torch.utils import data
import pickle
from matplotlib import pyplot as plt
import torch.utils.data as utils
import time
import os

bs = 2048
wd = 1e-2

is_cuda = torch.cuda.is_available()

class MultDataset(data.Dataset):
    def __init__(self, factors, product):
        'Initialization'
        self.factors = factors
        self.product = product

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.product)

    def __getitem__(self, index):
        # Load data and get label
        x = self.factors[index]
        y = self.product[index]

        return x, y

def rmse_loss(pred, targ):
    denom = targ**2
    denom = torch.sqrt(denom.sum()/len(denom))
    return torch.sqrt(F.mse_loss(pred, targ))/denom

def NN_train(pathdir, filename, epochs=1000, lrs=1e-2, N_red_lr=4, pretrained_path=""):
    try:
        os.mkdir("results/NN_trained_models/")
    except:
        pass

    try:
        os.mkdir("results/NN_trained_models/models/")
    except:
        pass
    try:
        n_variables = np.loadtxt(pathdir+"%s" %filename, dtype='str').shape[1]-1
        variables = np.loadtxt(pathdir+"%s" %filename, usecols=(0,))

        epochs = 200*n_variables
        if len(variables)<5000:
            epochs = epochs*3

        if n_variables==0 or n_variables==1:
            print("Solved!")#, variables[0])
            return 0

        else:
            for j in range(1,n_variables):
                v = np.loadtxt(pathdir+"%s" %filename, usecols=(j,))
                variables = np.column_stack((variables,v))

        f_dependent = np.loadtxt(pathdir+"%s" %filename, usecols=(n_variables,))
        f_dependent = np.reshape(f_dependent,(len(f_dependent),1))

        factors = torch.from_numpy(variables)
        if is_cuda:
            factors = factors.cuda()
        else:
            factors = factors
        factors = factors.float()

        product = torch.from_numpy(f_dependent)
        if is_cuda:
            product = product.cuda()
        else:
            product = product
        product = product.float()

        class SimpleNet(nn.Module):
            def __init__(self, ni):
                super().__init__()
                self.linear1 = nn.Linear(ni, 128)
                self.linear2 = nn.Linear(128, 128)
                self.linear3 = nn.Linear(128, 64)
                self.linear4 = nn.Linear(64,64)
                self.linear5 = nn.Linear(64,1)
            
            def forward(self, x):
                x = F.tanh(self.linear1(x))
                x = F.tanh(self.linear2(x))
                x = F.tanh(self.linear3(x))
                x = F.tanh(self.linear4(x))
                x = self.linear5(x)
                return x

        my_dataset = utils.TensorDataset(factors,product) # create your datset
        my_dataloader = utils.DataLoader(my_dataset, batch_size=bs, shuffle=True) # create your dataloader

        if is_cuda:
            model_feynman = SimpleNet(n_variables).cuda()
        else:
            model_feynman = SimpleNet(n_variables)

        if pretrained_path!="":
            model_feynman.load_state_dict(torch.load(pretrained_path))

        check_es_loss = 10000

        for i_i in range(N_red_lr):
            optimizer_feynman = optim.Adam(model_feynman.parameters(), lr = lrs)
            for epoch in range(epochs):
                model_feynman.train()
                for i, data in enumerate(my_dataloader):
                    optimizer_feynman.zero_grad()
                
                    if is_cuda:
                        fct = data[0].float().cuda()
                        prd = data[1].float().cuda()
                    else:
                        fct = data[0].float()
                        prd = data[1].float()
                    
                    loss = rmse_loss(model_feynman(fct),prd)
                    loss.backward()
                    optimizer_feynman.step()
                
                '''
                # Early stopping
                if epoch%20==0 and epoch>0:
                    if check_es_loss < loss:
                        break
                    else:
                        torch.save(model_feynman.state_dict(), "results/NN_trained_models/models/" + filename + ".h5")
                        check_es_loss = loss
                if epoch==0:
                    if check_es_loss < loss:
                        torch.save(model_feynman.state_dict(), "results/NN_trained_models/models/" + filename + ".h5")
                        check_es_loss = loss
                '''
                torch.save(model_feynman.state_dict(), "results/NN_trained_models/models/" + filename + ".h5")   
            print(loss)
            lrs = lrs/10

        return model_feynman

    except NameError:
        print("Error in file: %s" %filename)
        raise





import numpy as np
import matplotlib.pyplot as plt
import os
from os import path
from .get_pareto import Point, ParetoSet
from .RPN_to_pytorch import RPN_to_pytorch
from .RPN_to_eq import RPN_to_eq
from .S_NN_train import NN_train
from .S_NN_eval import NN_eval
from .S_symmetry import *
from .S_separability import *
from .S_change_output import *
from .S_brute_force import brute_force
from .S_combine_pareto import combine_pareto
from sympy.parsing.sympy_parser import parse_expr
from sympy import preorder_traversal, count_ops
from .S_polyfit import polyfit
from .S_get_symbolic_expr_error import get_symbolic_expr_error
from .S_add_snap_expr_on_pareto import add_snap_expr_on_pareto
from .S_add_sym_on_pareto import add_sym_on_pareto
from .S_run_bf_polyfit import run_bf_polyfit
from .S_final_gd import final_gd
from .S_add_bf_on_numbers_on_pareto import add_bf_on_numbers_on_pareto
from .dimensionalAnalysis import dimensionalAnalysis
from .S_NN_get_gradients import evaluate_derivatives
from .S_brute_force_comp import brute_force_comp
from .S_brute_force_gen_sym import brute_force_gen_sym
from .S_compositionality import *
from .S_gen_sym import *
from .S_gradient_decomposition import identify_decompositions

PA = ParetoSet()
def run_AI_all(pathdir,filename,BF_try_time=60,BF_ops_file_type="14ops", polyfit_deg=4, NN_epochs=4000, PA=PA):
    try:
        os.mkdir("results/")
    except:
        pass

    # load the data for different checks
    data = np.loadtxt(pathdir+filename)

    # Run bf and polyfit
    PA = run_bf_polyfit(pathdir,pathdir,filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_squared(pathdir,"results/mystery_world_squared/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)   

    # Run bf and polyfit on modified output

    PA = get_acos(pathdir,"results/mystery_world_acos/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_asin(pathdir,"results/mystery_world_asin/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_atan(pathdir,"results/mystery_world_atan/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_cos(pathdir,"results/mystery_world_cos/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_exp(pathdir,"results/mystery_world_exp/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_inverse(pathdir,"results/mystery_world_inverse/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_log(pathdir,"results/mystery_world_log/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_sin(pathdir,"results/mystery_world_sin/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_sqrt(pathdir,"results/mystery_world_sqrt/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_squared(pathdir,"results/mystery_world_squared/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)
    PA = get_tan(pathdir,"results/mystery_world_tan/",filename,BF_try_time,BF_ops_file_type, PA, polyfit_deg)

#############################################################################################################################
    # check if the NN is trained. If it is not, train it on the data.
    print("Checking for symmetry \n", filename)
    if len(data[0])<3:
        print("Just one variable!")
        pass
    elif path.exists("results/NN_trained_models/models/" + filename + ".h5"):# or len(data[0])<3:
        print("NN already trained \n")
        print("NN loss: ", NN_eval(pathdir,filename)[0], "\n")
        model_feynman = NN_eval(pathdir,filename)[1]
    elif path.exists("results/NN_trained_models/models/" + filename + "_pretrained.h5"):
        print("Found pretrained NN \n")
        model_feynman = NN_train(pathdir,filename,NN_epochs/2,lrs=1e-3,N_red_lr=3,pretrained_path="results/NN_trained_models/models/" + filename + "_pretrained.h5")
        print("NN loss after training: ", NN_eval(pathdir,filename), "\n")
    else:
        print("Training a NN on the data... \n")
        model_feynman = NN_train(pathdir,filename,NN_epochs)
        print("NN loss: ", NN_eval(pathdir,filename), "\n")

    # Check which symmetry/separability is the best
    # Symmetries
    symmetry_minus_result = check_translational_symmetry_minus(pathdir,filename)
    symmetry_divide_result = check_translational_symmetry_divide(pathdir,filename)
    symmetry_multiply_result = check_translational_symmetry_multiply(pathdir,filename)
    symmetry_plus_result = check_translational_symmetry_plus(pathdir,filename)

    # Separabilities
    separability_plus_result = check_separability_plus(pathdir,filename)
    separability_multiply_result = check_separability_multiply(pathdir,filename)

    print(symmetry_minus_result)
    print(symmetry_divide_result)
    print(symmetry_multiply_result)
    print(symmetry_plus_result)
    print(separability_plus_result)
    print(separability_multiply_result)

    if symmetry_plus_result[0]==-1:
        idx_min = -1
    else:
        idx_min = np.argmin(np.array([symmetry_plus_result[0], symmetry_minus_result[0], symmetry_multiply_result[0], symmetry_divide_result[0], separability_plus_result[0], separability_multiply_result[0]]))

    print("idx_min: ", idx_min)

    # Check if compositionality is better than the best so far
    if idx_min==0:
        mu, sigma = symmetry_plus_result[3:]
    elif idx_min==1:
        mu, sigma = symmetry_minus_result[3:]
    elif idx_min==2:
        mu, sigma = symmetry_multiply_result[3:]
    elif idx_min==3:
        mu, sigma = symmetry_divide_result[3:]
    elif idx_min==4:
        mu, sigma = separability_plus_result[3:]
    elif idx_min==5:
        mu, sigma = separability_multiply_result[3:]



    try:
        print("mu, sigma: ", mu, sigma)
    except:
        pass

    # Save the gradients for compositionality
    try:
        succ_grad = evaluate_derivatives(pathdir,filename,model_feynman)
    except:
        succ_grad = 0

    idx_comp = 0
    print("succ_grad: ", succ_grad)
    if succ_grad == 1:
        #try:
        for qqqq in range(1):
            brute_force_comp("results/","gradients_comp_%s.txt" %filename,600,"14ops.txt")
            bf_all_output = np.loadtxt("results_comp.dat", dtype="str")
            for bf_i in range(len(bf_all_output)):
                idx_comp_temp = 0
                try:
                    express = bf_all_output[:,1][bf_i]
                    idx_comp_temp, eqq, new_mu, new_sigma = check_compositionality(pathdir,filename,model_feynman,express,mu,sigma,nu=10)
                    print("check sigma")
                    print(idx_comp_temp, eqq, new_mu, new_sigma)
                    if idx_comp_temp==1:
                        print("yesssss")
                        idx_comp = 1
                        math_eq_comp = eqq
                        mu = new_mu
                        sigma = new_sigma
                except:
                    continue
        #except:
        #    idx_comp = 0
    else:
        idx_comp = 0

    print("idx_comp: ", idx_comp)

    if idx_comp==1:
        idx_min = 6



    # Check if generalized separabilty is better than the best so far
    idx_gen_sym = 0
    for kiiii in range(1):
        print("doing gen sym")
        if len(data[0])>3:
            # find the best separability indices
            decomp_idx = identify_decompositions(pathdir,filename, model_feynman)
            print("IDXXXXXXXXX ", decomp_idx)
            print("doing gen sym 1")
            brute_force_gen_sym("results/","gradients_gen_sym_%s" %filename,600,"14ops.txt")
            bf_all_output = np.loadtxt("results_gen_sym.dat", dtype="str")
            print("alllllllllllll")
            print(bf_all_output)

            for bf_i in range(len(bf_all_output)):
                idx_gen_sym_temp = 0
                try:
                    express = bf_all_output[:,1][bf_i]
                    idx_gen_sym_temp, eqq, new_mu, new_sigma = check_gen_sym(pathdir,filename,model_feynman,decomp_idx,express,mu,sigma,nu=10)
                    print("check sigma")
                    print(idx_gen_sym_temp, eqq, new_mu, new_sigma)
                    if idx_gen_sym_temp==1:
                        idx_gen_sym = 1
                        math_eq_gen_sym = eqq
                        mu = new_mu
                        sigma = new_sigma
                except:
                    continue

    print("idx_gen_sym: ", idx_gen_sym)

    if idx_gen_sym==1:
        idx_min = 7


    # Apply the best symmetry/separability and rerun the main function on this new file
    if idx_min == 0:
        new_pathdir, new_filename = do_translational_symmetry_plus(pathdir,filename,symmetry_plus_result[1],symmetry_plus_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA = add_sym_on_pareto(pathdir,filename,PA1,symmetry_plus_result[1],symmetry_plus_result[2],PA,"+")
        return PA

    elif idx_min == 1:
        new_pathdir, new_filename = do_translational_symmetry_minus(pathdir,filename,symmetry_minus_result[1],symmetry_minus_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA = add_sym_on_pareto(pathdir,filename,PA1,symmetry_minus_result[1],symmetry_minus_result[2],PA,"-")
        return PA

    elif idx_min == 2:
        new_pathdir, new_filename = do_translational_symmetry_multiply(pathdir,filename,symmetry_multiply_result[1],symmetry_multiply_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA = add_sym_on_pareto(pathdir,filename,PA1,symmetry_multiply_result[1],symmetry_multiply_result[2],PA,"*")
        return PA

    elif idx_min == 3:
        new_pathdir, new_filename = do_translational_symmetry_divide(pathdir,filename,symmetry_divide_result[1],symmetry_divide_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA = add_sym_on_pareto(pathdir,filename,PA1,symmetry_divide_result[1],symmetry_divide_result[2],PA,"/")
        return PA

    elif idx_min == 4:
        new_pathdir1, new_filename1, new_pathdir2, new_filename2,  = do_separability_plus(pathdir,filename,separability_plus_result[1],separability_plus_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir1,new_filename1,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA2_ = ParetoSet()
        PA2 = run_AI_all(new_pathdir2,new_filename2,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA2_)
        combine_pareto_data = np.loadtxt(pathdir+filename)
        PA = combine_pareto(combine_pareto_data,PA1,PA2,separability_plus_result[1],separability_plus_result[2],PA,"+")
        return PA

    elif idx_min == 5:
        new_pathdir1, new_filename1, new_pathdir2, new_filename2,  = do_separability_multiply(pathdir,filename,separability_multiply_result[1],separability_multiply_result[2])
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir1,new_filename1,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA2_ = ParetoSet()
        PA2 = run_AI_all(new_pathdir2,new_filename2,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA2_)
        combine_pareto_data = np.loadtxt(pathdir+filename)
        PA = combine_pareto(combine_pareto_data,PA1,PA2,separability_multiply_result[1],separability_multiply_result[2],PA,"*")
        return PA

    elif idx_min == 6:
        new_pathdir, new_filename = do_compositionality(pathdir,filename,math_eq_comp)
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        print("check vvv")
        PA = add_comp_on_pareto(PA1,PA,math_eq_comp)
        return PA

    elif idx_min == 7:
        new_pathdir, new_filename = do_gen_sym(pathdir,filename,decomp_idx,math_eq_gen_sym)
        PA1_ = ParetoSet()
        PA1 = run_AI_all(new_pathdir,new_filename,BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA1_)
        PA = add_gen_sym_on_pareto(PA1,PA, decomp_idx, math_eq_gen_sym)
        return PA
    else:
        print("FINAL FINAL")
        return PA
# this runs snap on the output of aifeynman
def run_aifeynman(pathdir,filename,BF_try_time,BF_ops_file_type, polyfit_deg=4, NN_epochs=4000, vars_name=[],test_percentage=20):
    # If the variable names are passed, do the dimensional analysis first
    filename_orig = filename
    try:
        if vars_name!=[]:
            dimensionalAnalysis(pathdir,filename,vars_name)
            DR_file = filename + "_dim_red_variables.txt"
            filename = filename + "_dim_red"
        else:
            DR_file = ""
    except:
        DR_file = ""

    # Split the data into train and test set
    input_data = np.loadtxt(pathdir+filename)
    sep_idx = np.random.permutation(len(input_data))

    train_data = input_data[sep_idx[0:(100-test_percentage)*len(input_data)//100]]
    test_data = input_data[sep_idx[test_percentage*len(input_data)//100:len(input_data)]]

    np.savetxt(pathdir+filename+"_train",train_data)
    if test_data.size != 0:
        np.savetxt(pathdir+filename+"_test",test_data)

    PA = ParetoSet()
    # Run the code on the train data
    PA = run_AI_all(pathdir,filename+"_train",BF_try_time,BF_ops_file_type, polyfit_deg, NN_epochs, PA=PA)
    PA_list = PA.get_pareto_points()

    '''
    # Run bf snap on the resulted equations
    for i in range(len(PA_list)):
        try:
            PA = add_bf_on_numbers_on_pareto(pathdir,filename,PA,PA_list[i][-1])
        except:
            continue
    PA_list = PA.get_pareto_points()
    '''

    np.savetxt("results/solution_before_snap_%s.txt" %filename,PA_list,fmt="%s")


    # Run zero, integer and rational snap on the resulted equations
    for j in range(len(PA_list)):
        PA = add_snap_expr_on_pareto(pathdir,filename,PA_list[j][-1],PA, "")

    PA_list = PA.get_pareto_points()
    np.savetxt("results/solution_first_snap_%s.txt" %filename,PA_list,fmt="%s")

    # Run gradient descent on the data one more time
    for i in range(len(PA_list)):
        try:
            gd_update = final_gd(pathdir,filename,PA_list[i][-1])
            PA.add(Point(x=gd_update[1],y=gd_update[0],data=gd_update[2]))
        except:
            continue

    PA_list = PA.get_pareto_points()
    for j in range(len(PA_list)):
        PA = add_snap_expr_on_pareto(pathdir,filename,PA_list[j][-1],PA, DR_file)

    list_dt = np.array(PA.get_pareto_points())
    data_file_len = len(np.loadtxt(pathdir+filename))
    log_err = []
    log_err_all = []
    for i in range(len(list_dt)):
        log_err = log_err + [np.log2(float(list_dt[i][1]))]
        log_err_all = log_err_all + [data_file_len*np.log2(float(list_dt[i][1]))]
    log_err = np.array(log_err)
    log_err_all = np.array(log_err_all)

    # Try the found expressions on the test data
    if DR_file=="" and test_data.size != 0:
        test_errors = []
        input_test_data = np.loadtxt(pathdir+filename+"_test")
        for i in range(len(list_dt)):
            test_errors = test_errors + [get_symbolic_expr_error(input_test_data,str(list_dt[i][-1]))]
        test_errors = np.array(test_errors)
        # Save all the data to file
        save_data = np.column_stack((test_errors,log_err,log_err_all,list_dt))
    else:
        save_data = np.column_stack((log_err,log_err_all,list_dt))
    np.savetxt("results/solution_%s" %filename_orig,save_data,fmt="%s")
    try:
        os.remove(pathdir+filename+"_test")
        os.remove(pathdir+filename+"_train")
    except:
        pass




import setuptools
from numpy.distutils.core import Extension, setup

sr1 = Extension(name='aifeynman._symbolic_regress1', sources=[
    'aifeynman/symbolic_regress1.f90'])

sr2 = Extension(name='aifeynman._symbolic_regress2', sources=[
    'aifeynman/symbolic_regress2.f90'])

sr3 = Extension(name='aifeynman._symbolic_regress3', sources=[
    'aifeynman/symbolic_regress3.f90'])

sr_mdl_mult = Extension(name='aifeynman._symbolic_regress_mdl3', sources=[
    'aifeynman/symbolic_regress_mdl3.f90'])

sr_mdl_plus = Extension(name='aifeynman._symbolic_regress_mdl2', sources=[
    'aifeynman/symbolic_regress_mdl2.f90'])

sr_mdl4 = Extension(name='aifeynman._symbolic_regress_mdl4', sources=[
    'aifeynman/symbolic_regress_mdl4.f90'])

sr_mdl5 = Extension(name='aifeynman._symbolic_regress_mdl5', sources=[
    'aifeynman/symbolic_regress_mdl5.f90'])

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='aifeynman',
    version='2.0.6',
    description='AI Feynman: a Physics-Inspired Method for Symbolic Regression',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/SJ001/aifeynman',
    author='Silviu-Marian Udrescu',
    author_email='sudrescu@mit.edu',
    license='MIT',
    packages=['aifeynman'],
    package_dir={'aifeynman': 'aifeynman'},
    package_data={'aifeynman': ['*.txt', '*.f90']},
    include_package_data=True,
    ext_modules=[sr1, sr2, sr3, sr_mdl_mult, sr_mdl_plus, sr_mdl4, sr_mdl5],
    python_requires='>3.6',
    install_requires=['matplotlib',
                      'numpy',
                      'seaborn',
                      'sklearn',
                      'sortedcontainers',
                      'sympy >= 1.4',
                      'torch >= 1.4.0',
                      'torchvision',
                      ],
    entry_points={
        'console_scripts': [
            'feynman_sr1 = aifeynman._symbolic_regress1:go',
            'feynman_sr2 = aifeynman._symbolic_regress2:go',
            'feynman_sr3 = aifeynman._symbolic_regress3:go',
            'feynman_sr_mdl_mult = aifeynman._symbolic_regress_mdl3:go',
            'feynman_sr_mdl_plus = aifeynman._symbolic_regress_mdl2:go',
            'feynman_sr_mdl4 = aifeynman._symbolic_regress_mdl4:go',
            'feynman_sr_mdl5 = aifeynman._symbolic_regress_mdl5:go',
        ]}
)
